/*******************************************************************************
* Licensed Materials - Property of HCL
* (c) Copyright HCL Technologies Ltd. 2016, 2018.  All Rights Reserved.
*
* Note to U.S. Government Users Restricted Rights:
* Use, duplication or disclosure restricted by GSA ADP Schedule
* Contract with IBM Corp.
*******************************************************************************/

/**
 * Diagram viewer
 *
 * Note: In this file we cannot use "modern" JavaScript since it will be loaded in Jazz webpages which have
 * ===== set compatibility mode to older versions of Internet Explorer where such JavaScript cannot be used!
 *
 * @author Mattias Mohlin
 */

define([
    'jquery-ui',
    'svg-pan-zoom',
    'utils',    
    'menu'],
function (jquery_ui, svgPanZoom, Utils, Menu) {

    var VIEW_NAME = 'diagramViewerComponent_';

    function DiagramViewer() { }    

    var zoom_values = [
        0.10,
        0.25,
        0.50,
        0.75,
        1.00,
        1.50,
        2.00,
        2.50,
        3.00,
        4.00,
        5.00
    ];

    var default_zoom_index = zoom_values.indexOf(1.00);
    var max_zoom_index = zoom_values.length - 1;   

    // An instance of Diagram below is created for each open diagram
    function Diagram(diagramViewerDiv) {
        this.diagramViewerDiv = diagramViewerDiv;        
        this.panZoomer = null; // Instance of svg-pan-zoom used for panning and zooming the diagram
        this.currentZoomIndex = default_zoom_index;     
        this.aboutToClickInBackground = false; // Flag to distinguish between click in background (for symbol deselection) and panning        

        /**
         * Initialize the diagram toolbar          
         */
        this.initToolbar = function () {
            var self = this;           

            $(self.diagramViewerDiv).find('.zoom_in_button').click(function () {
                self.zoomIn();
            });
            $(self.diagramViewerDiv).find('.zoom_out_button').click(function () {
                self.zoomOut();
            });
            $(self.diagramViewerDiv).find('.center_button').click(function () {
            //  self.panZoomer.center();
            	location.reload();                
            });                        

            // All browsers behave slightly differently in how they position the zoom slider. Adjust its vertical offset
            // depending on browser.
            if (Utils.isFirefox()) { 
                $(self.diagramViewerDiv).find('.zoom_slider').css('top', '8px'); 
            }
            else if (Utils.isChrome()) {
                $(self.diagramViewerDiv).find('.zoom_slider').css('top', '3px'); 
            }
            // Default offset of 0 works for IE

            $(self.diagramViewerDiv).find('.zoom_slider').slider({
                min: 0,
                max: max_zoom_index + 1, // Add one extra value which cannot be set, just to make the slider handler better align with the slider line        
                slide: function (event, ui) {
                    if (ui.value == max_zoom_index + 1)
                        return false;
                    self.setZoom(ui.value);                        
                }
            });
            $(self.diagramViewerDiv).find(".zoom_slider .ui-slider-handle").unbind('keydown'); // disable arrow keys for moving slider since they are used for diagram panning

            $(self.diagramViewerDiv).find('.zoom_level_text').change(function (event) {                
                var newZoom = $(event.target).val().trim();
                if (newZoom.length > 0 && newZoom[newZoom.length - 1] == '%')
                    newZoom = newZoom.substring(0, newZoom.length - 1);
                var newZoomValue = parseInt(newZoom);
                if (newZoomValue) {
                    var newZoomFactor = newZoomValue / 100;
                    if (newZoomFactor < zoom_values[0]) {
                        newZoomFactor = zoom_values[0];
                    }
                    else if (newZoomFactor > zoom_values[zoom_values.length - 1]) {
                        newZoomFactor = zoom_values[zoom_values.length - 1];                        
                    }
                    self.updateZoomUI(newZoomFactor); // Needed in case zoom does not happen (when entering values outside supported zoom range)
                    self.panZoomer.zoom(newZoomFactor);
                    $(this).blur();
                }
                else {
                    // Not a valid number. Print current zoom level.
                    var currentZoom = Math.round(self.panZoomer.getZoom() * 100);
                    $(this).val(currentZoom + '%');
                }
            });
            $(self.diagramViewerDiv).find('.zoom_level_text').keydown(function (event) {   
                if (event.which == 13 /* Enter */ || event.which == 27 /* Escape */) {
                    $(this).blur();
                }
                event.stopPropagation(); // Prevent other keystrokes in the zoom field to be interpreted as keyboard shortcuts
                // Only allow numeric input and some special characters
                if (["Backspace", "Delete", "ArrowUp", "Up", "ArrowLeft", "Left", "ArrowDown", "Down", "ArrowRight", "Right", "Home", "End",
                    "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"].indexOf(event.key) > -1) {
                    return true;
                }
                return false;
                
            });

            self.updateZoomUI();
        };
        
        /**
         * Initialize the diagram graphics            
         */
        this.initGraphics = function () {
            var self = this;

            // TODO: Call here a Java method which will return the SVG as a string.
            // Then uncomment the line below to insert it into the div.
            // $(self.diagramViewerDiv).append(svg);
			$(self.diagramViewerDiv).append(javaFunctionLoadSvgData());
			
			$("a").click(function(event) {
				//block click option
				return false
							
			});
			
			$("a").dblclick(function(event) {
									var href = $(this).attr('xlink:href');
									//alert(href);
									openTCJSFile(href.toString())
									event.preventDefault();
			});
					
            var svgRoot = $(self.diagramViewerDiv).find('svg');
            self.panZoomer = svgPanZoom(svgRoot[0]);             

            svgRoot.css({
                width: $(self.diagramViewerDiv).width(),
                height: $(self.diagramViewerDiv).height()
            });

            svgRoot.addClass('grabbable'); // Should ideally change to 'grabbing' while panning                        

            // Add support for selecting and deselecting symbols in the diagram
            svgRoot.mousedown(function (e) {
                if (!e.originalEvent.originatedOnShape)
                    self.aboutToClickInBackground = true;
            });
            svgRoot.mouseup(function (e) {
                if (self.aboutToClickInBackground) {
                    if (!e.ctrlKey)
                        self.deselectAllSymbols();
                    self.aboutToClickInBackground = false;
                }
            });

            self.deselectAllSymbols(); // Start with empty selection (diagram itself becomes implicit selection)            

            // Select a shape in the diagram
            function selectInDiagram(shape, event) {
                var elementId = shape.data('model-element');
                if (!elementId) {
                    // Selected shape has no model element. Reset current selection and don't make a new one.
                    self.deselectAllSymbols();
                    return;
                }
                
                shape.removeClass('can-be-selected-on-diagram');
                
                var newSelection = [{                    
                    'url': 'TODO: TC URL here'
                }];
                if (!event.ctrlKey) {
                    self.deselectAllSymbols(true /* newSelectionToBeSet */);
                    shape.addClass('selected-on-diagram');
                }              
            }

            $(self.diagramViewerDiv).find('.selectable').click(function (e) {                  
                selectInDiagram($(this), e);
                e.stopPropagation();
            });
            $(self.diagramViewerDiv).find('.selectable').mouseenter(function (e) {
                if (!$(this).hasClass('selected-on-diagram')) {
                    $(this).addClass('can-be-selected-on-diagram');
                }
            });
            $(self.diagramViewerDiv).find('.selectable').mouseleave(function (e) {
                $(this).removeClass('can-be-selected-on-diagram');
            });            
            $(self.diagramViewerDiv).find('.selectable').mousedown(function (e) {
                self.aboutToClickInBackground = false;
                e.originalEvent.originatedOnShape = true; // Store the fact that this event originated on a shape (this info is needed in the mousedown event handler on the SVG root - see above)
            });

            // Context menu
            $(self.diagramViewerDiv).find('.selectable').contextmenu(function (e) {
                var svgShape = $(this);
                if (!svgShape.hasClass('selected-on-diagram')) {
                    selectInDiagram(svgShape, e); // Bringing up the context menu automatically selects the shape
                }                

                var contextMenu = new Menu('diagram_context_menu' /* id */, $('#dynamic_container'));
                contextMenu.addMenuItem('cmd_show_in_explorer', '{{show_in_explorer}}', {});                

                contextMenu.onMenuItemClicked(function (id) {
                    if (id == 'cmd_show_in_explorer') {
                        var qualifiedId = self.currentDesign.id + '/' + svgShape.data('model-element') + '/model';
                        //view_manager.showInExplorerView(qualifiedId);
                    }                    
                });

                contextMenu.enableKeyboardShortcuts();
                contextMenu.open({ my: "left top", at: "center", of: e });                

                return false; // Prevent default browser context menu
            });

            // Navigation to nested diagram by double-click on symbol
            $(self.diagramViewerDiv).find('.selectable').dblclick(function (e) {
                var svgShape = $(this);
                var qualifiedId = self.currentDesign.id + '/' + svgShape.data('model-element') + '/model';
                //view_manager.openContainedDiagram(qualifiedId);
                e.stopPropagation();
            });

            self.panZoomer.setBeforePan(function () {
                self.aboutToClickInBackground = false;
            });

            self.panZoomer.disableDblClickZoom();
            self.panZoomer.setMinZoom(zoom_values[0]);
            self.panZoomer.setMaxZoom(zoom_values[zoom_values.length - 1]);            
            self.panZoomer.setZoomScaleSensitivity(0.25); // When mouse scroll-wheel is used
            self.setZoom(self.currentZoomIndex);

            self.panZoomer.setOnZoom(function (zoomFactor) {               
                // Handle case of custom zooming (scroll wheel)

                // Adjust current zoom index to be as close as possible to this custom zoom factor
                var minDiff = 100;                
                for (var i = 0; i < zoom_values.length; i++) {
                    var diff = Math.abs(zoomFactor - zoom_values[i]);
                    if (diff < minDiff) {
                        minDiff = diff;
                        self.currentZoomIndex = i;
                    }
                    if (diff == 0)
                        break; // Best possible match - look no further
                }
            
                self.updateZoomUI(zoomFactor);                
            });            

        }        

        /**
         * Deselect all symbols in the diagram (i.e reset the selection).
         * @param {boolean} newSelectionToBeSet Set this parameter to true if a new selection will be set immediately after resetting the current one.
         *                                      If this parameter is not true, the diagram itself will be set as the implicit selection.
         */
        this.deselectAllSymbols = function (newSelectionToBeSet) {
            var self = this;

            $(self.diagramViewerDiv).find('.selectable').removeClass('selected-on-diagram');

            if (newSelectionToBeSet)
                return;            
        }
    
        /**
        * Refresh the zoom UI in the toolbar according to current zoom level.        
        * If a customZoomFactor is provided (case of mouse wheel zooming, or entering a custom zoom level in the field)
        * it is used as the zoom percentage text
        */
        this.updateZoomUI = function (customZoomFactor) {
            var self = this;

            $(self.diagramViewerDiv).find('.zoom_in_button').prop("disabled", false);
            $(self.diagramViewerDiv).find('.zoom_out_button').prop("disabled", false);

            if (self.currentZoomIndex == max_zoom_index)
                $(self.diagramViewerDiv).find('.zoom_in_button').prop("disabled", true);
            if (self.currentZoomIndex == 0)
                $(self.diagramViewerDiv).find('.zoom_out_button').prop("disabled", true);
               
            var slider = $(self.diagramViewerDiv).find('.zoom_slider');
            slider.slider('value', self.currentZoomIndex);
            var value = slider.slider( "option", "value" );
            var zoomPercentage = customZoomFactor ? Math.round(customZoomFactor * 100) : zoom_values[self.currentZoomIndex] * 100;
            $(self.diagramViewerDiv).find('.zoom_level_text').val(zoomPercentage + '%');
        }


        /**
         * Increase zoom level
         */
        this.zoomIn = function () {
            var self = this;            
            
            var current = self.currentZoomIndex;

            current += 1;
            if (current > max_zoom_index) {
                // set max
                current = max_zoom_index;
            }

            self.setZoom(current);
        }
        
        /**
         * Decrease zoom level
         */
        this.zoomOut = function () {
            var self = this;            
            
            var current = self.currentZoomIndex;

            current -= 1;
            if (current < 0) {
                // set min
                current = 0;
            }

            self.setZoom(current);
        }        

        /**
         * Zoom to the specified zoom level
         * @param {int} index New zoom level to set
        */
        this.setZoom = function(index) {            
            var self = this;

            self.panZoomer.zoom(zoom_values[index]);
            self.currentZoomIndex = index;
            self.updateZoomUI();
        }

        // Enable keybindings for the diagram
        this.activateKeybindings = function () {
            var self = this;

            // Start by ensuring that our keybindings are not already enabled (avoid enabling them twice)
            $(document).off('.' + self.getViewName());

            $(document).on('keydown.' + VIEW_NAME + this.diagramId, function (e) {
                switch (e.key) {
                    case "+":
                        self.zoomIn();
                        break;
                    case "-":
                        self.zoomOut();
                        break;
                    case "Home":
                        self.panZoomer.center();
                        break;
                    case "ArrowUp":
                    case "Up": // IE
                        self.panZoomer.panBy({ x: 0, y: -10 })
                        break;
                    case "ArrowDown":
                    case "Down": // IE
                        self.panZoomer.panBy({ x: 0, y: 10 })
                        break;
                    case "ArrowLeft":
                    case "Left": // IE
                        self.panZoomer.panBy({ x: -10, y: 0 })
                        break;
                    case "ArrowRight":
                    case "Right": // IE
                        self.panZoomer.panBy({ x: 10, y: 0 })
                        break;
                }
            });                            
        }

        function getIdFromView(view) {
            if (!view)
                return null;

            var i = view.lastIndexOf(VIEW_NAME, 0);
            if (i != 0)
                return null;

            return view.substring(i + VIEW_NAME.length);
        }        

        // Return the (static) name of this view
        this.getViewName = function () {
            return VIEW_NAME + this.diagramId;
        }

        /**
         * Should be called when the diagram container is resized so that the diagram can adjust
         * its contents accordingly.
         * @param {int} containerWidth New width of container
         * @param {int} containerHeight New height of container
         */
        this.onContainerResized = function (containerWidth, containerHeight) {
            var self = this;

            $(self.diagramViewerDiv).css({
                width: containerWidth,
                height: containerHeight
            });
            $(self.diagramViewerDiv).find('svg').css({
                width: containerWidth,
                height: containerHeight
            });
            
            // Since we are not using the floating zoom controls from the svg-pan-zoom library, we don't need to
            // resize the panZoomer. Doing so has the drawback of making the zoom levels dependent on the container size.            
        }

        /**
         * Should be called when the diagram container is resized so that the diagram can adjust
         * its contents accordingly.
         * @param {int} containerWidth New width of container
         * @param {int} containerHeight New height of container
         */
        this.onContainerResized = function (containerWidth, containerHeight) {
            var self = this;

            $(self.diagramViewerDiv).css({
                width: containerWidth,
                height: containerHeight
            });
            $(self.diagramViewerDiv).find('svg').css({
                width: containerWidth,
                height: containerHeight
            });
            
            // Since we are not using the floating zoom controls from the svg-pan-zoom library, we don't need to
            // resize the panZoomer. Doing so has the drawback of making the zoom levels dependent on the container size.            
        }

        // Select the shape on the diagram that shows the element with the specified id. If multiple such shapes exist
        // use the symbolIndex to decide which one to select. Also pan the diagram so the selected shape becomes visible.
        this.selectShapeShowingElement = function (id, symbolIndex) {
            var self = this;

            var shapesShowingElement = $(self.diagramViewerDiv).find('[data-model-element="' + id + '"]');
            var selectedShape;
            if (shapesShowingElement.length == 1 || !symbolIndex) {
                selectedShape = shapesShowingElement[0];                
            }
            else {
                // There are more than one shape on the diagram that shows the element. Use the symbolIndex
                // query parameter to determine which one to select.                
                if (symbolIndex < 0 || symbolIndex > (shapesShowingElement.length - 1)) {
                    // Erroneous symbol index. Just select any of them.
                    selectedShape = shapesShowingElement[0];                    
                }
                else {
                    selectedShape = shapesShowingElement[symbolIndex]                    
                }
            }
            $(selectedShape).click();

            // TODO: For now we don't call resize() when the container is resized so get the width and height from the svg element instead
            // of from sizes.
            // Calling resize() may have some benefits but also gave some unwanted effects (see comment for onContainerResized above).
            var sizes = self.panZoomer.getSizes();

            var w = $(self.diagramViewerDiv).find('svg').width();
            var h = $(self.diagramViewerDiv).find('svg').height();
            var x = parseInt($(selectedShape).attr('x'));
            var y = parseInt($(selectedShape).attr('y'));

            self.panZoomer.pan({
                x: -(x * sizes.realZoom) + (w / 2),
                y: -(y * sizes.realZoom) + (h / 2)
            });
        }

        /**
         * Called when an error occurs. The error message will be shown in the diagram area
         * and the toolbar will be disabled.
         * @param {string} msg Error message (may be HTML)
         */
        this.error = function (msg) {
            $(this.diagramViewerDiv).find('.zoom_in_button').prop("disabled", true);
            $(this.diagramViewerDiv).find('.zoom_out_button').prop("disabled", true);
            $(this.diagramViewerDiv).find('.zoom_selector').prop("disabled", true);
            $(this.diagramViewerDiv).find('.diagram').html(msg);
        }
    }

    /**
     * Open a new diagram inside the diagramViewerDiv
     * @param {div} diagramViewerDiv A diagramViewer div where to place the new diagram     
     * @param {function} callback If specified, this function will be called (with the diagram as argument) when the diagram is open.
     */    
    DiagramViewer.prototype.open = function (diagramViewerDiv, callback) {
    	var self = this;        

        var diagram = new Diagram(diagramViewerDiv);

        diagram.initToolbar();    
                    
        diagram.initGraphics();

        // Store the diagram object on the diagramViewerDiv so that we can interact
        // with it later when needed.
        diagramViewerDiv.diagram = diagram;

        $(diagramViewerDiv).attr('tabindex', "-1");
        $(diagramViewerDiv).focus();

        diagram.activateKeybindings();                

        if (callback)
            callback(diagram);
    }                            	

    return new DiagramViewer();
});
