/*******************************************************************************
* Licensed Materials - Property of HCL
* (c) Copyright HCL Technologies Ltd. 2017, 2018.  All Rights Reserved.
*
* Note to U.S. Government Users Restricted Rights:
* Use, duplication or disclosure restricted by GSA ADP Schedule
* Contract with IBM Corp.
*******************************************************************************/

/**
 * Popup menu implementation based on jQueryUI menus
 *
 * Note: In this file we cannot use "modern" JavaScript since it will be loaded in Jazz webpages which have
 * ===== set compatibility mode to older versions of Internet Explorer where such JavaScript cannot be used!
 *
 * @author Mattias Mohlin
 */

define(['utils'], function (Utils) {    

    var currentlyOpenMenu = null; // The currently open menu (at most one can be open at the same time)

    /**
     * Creates an empty popup menu ready to be populated. The menu can be created either from an existing DOM element
     * or as a new DOM element.
     *
     * Create from existing DOM element:
     * @param {element} menuElement DOM element with class dropdown-menu to create the menu from
     *
     * Create new DOM element:
     * @param {string} id Unique menu DOM id
     * @param {element} container Container DOM element where to add the menu
     */
    function Menu(arg1, container) {
        this.menu = null;                   // jQueryUI menu object
        this.fromExistingDOMElement = true; // Is the menu created from an existing DOM element?
        this.hasBeenOpened = false;         // Has this menu been opened yet?

        if ((typeof arg1) == 'string') {
            var id = arg1;            
            var menuHTML = '<ul class="dropdown-menu" id="' + id + '"></ul>';            

            container.append($(menuHTML));
            this.menu = container.find('.dropdown-menu');
            this.fromExistingDOMElement = false;
        }
        else {
            this.menu = arg1;
        }
    }

    /**
     * Adds a menu item to the menu
     *     
     * @param {string} id Unique menu item DOM id
     * @param {string} label Menu item label (must not be empty)
     * @param {object} opt Optional settings for the menu item
     * @param {boolean} opt.enabled If false, the menu item will be disabled (default true)  
     * @param {boolean} opt.active If true, the menu item will be marked with a checkmark (default false)
     * @param {boolean} opt.dynamic Set to true for a dynamic menu item. In this case the id is set as a data attribute on the DOM element instead. (default false)
     * @param {string} opt.iconURL Specifies the icon of the menu item using an image URL
     * @param {string} opt.iconDivId Specifies the icon of the menu item using a div with a specific id
     * @param {boolean} opt.escapeHTML Should the label be HTML-escaped? If false, caller is responsible to HTML-escape the label. (default true)
     * @returns {jQueryElement} The added menu item
     */
    Menu.prototype.addMenuItem = function (id, label, opt) {
        Utils.mergeProperties(opt, {
            enabled: true,
            active: false,
            dynamic: false,
            iconURL: null,
            iconDivId: null,
            escapeHTML: true
        });        

        if (label == '')
            label = '"empty"'; // Label must not be empty

        var menuItemHTML = '<li ';
        var li_classes = '';
        if (opt.dynamic) {
            li_classes += 'dynamic-menu-item';
        }
        if (!opt.enabled) {
            li_classes += 'ui-state-disabled';
        }
        if (li_classes.length > 0) {
            menuItemHTML += 'class="' + li_classes + '"';
        }
        menuItemHTML += '>';
        if (opt.iconURL) {
            menuItemHTML += '<div class="ui-menu-item-wrapper"><img class="menu-item-icon" src="' + opt.iconURL + '"/>'
        }
        else if (opt.iconDivId) {
            menuItemHTML += '<div class="ui-menu-item-wrapper"><div id="' + opt.iconDivId + '" class="menu-item-icon"/>'
        }
        else {
            menuItemHTML += '<div>';
        }
        if (opt.active)
            menuItemHTML += '<span class="ui-icon ui-icon-check"/>';
        menuItemHTML += '<a ';
        if (!opt.dynamic) {
            menuItemHTML += 'id="' + id + '"';
        }
        else {
            menuItemHTML += 'data-id="' + id + '"';
        }
        menuItemHTML += '>' + (opt.escapeHTML ? Utils.escapeHtml(label) : label) + '</a></div></li>';
        var menuItem = $(menuItemHTML);
        this.menu.append(menuItem);

        return menuItem;
    };

    /**
     * Adds a separator to the menu.
     * @param {boolean} dynamic If true the separator will be treated like a dynamic menu item (i.e. removed automatically when closing the menu)
     */
    Menu.prototype.addSeparator = function (dynamic) {
        var htmlClasses = 'divider';
        if (dynamic)
            htmlClasses += ' dynamic-menu-item';
        this.menu.append($('<li class="' + htmlClasses + '"></li>'));
    };

    /**
     * Registers a callback for when a menu item is clicked. The callback function has two arguments: the id of the clicked menu item, and the
     * click event. It sets the 'this' context to the menu item itself.
     * This function should be called when all menu items have been added to the menu.
     */
    Menu.prototype.onMenuItemClicked = function (callback) {
        var self = this;

        self.menu.find('a').off('click').click(function (e) {
            var id = ($(e.currentTarget).parents('.dynamic-menu-item').length > 0) ?
                $(e.currentTarget).data('id') /* dynamic menu item */ :
                $(e.currentTarget).attr('id') /* regular menu item */;

            if (id)
                callback.apply($(e.currentTarget), [id, e]);            

            e.stopPropagation();
            self.close();
        });
    }

    /**
     * Registers a callback for when the menu is closed
     */
    Menu.prototype.onMenuClosed = function (callback) {
        this.menuClosedCallback = callback;
    }

    /**
     * Close the menu
     */
    Menu.prototype.close = function () {
        var self = this;

        if (currentlyOpenMenu != self)
            return; // This menu is not open so it cannot be closed

        $(document).off('mousedown.popupmenu');        

        if (!self.fromExistingDOMElement) {
            // Dynamically created menu: remove entire menu 
            self.menu.remove();
        }
        else {
            // Static DOM menu: only remove dynamically added menu items
            self.menu.find(".dynamic-menu-item").remove();
            self.menu.hide();
        }

        if (self.menuClosedCallback)
            self.menuClosedCallback();

        currentlyOpenMenu = null;
    }    

    /**
     * Opens the menu. Should be called after the menu has been fully initialized by other calls.
     * @param {object} position Optional jQueryUI position object specifying the position of the menu. If not specified the menu is positioned according to its DOM location.
     */
    Menu.prototype.open = function (position) {
        var self = this;

        // Only one menu can be open at the same time so close the currently open menu if any
        if (currentlyOpenMenu)
            currentlyOpenMenu.close();

        // Enable automatic close of the menu when clicked outside of it
        self.menu.on('mousedown', function (e) {
            e.stopPropagation();
        });
        $(document).on('mousedown.popupmenu', function () {
            self.close();
        });

        // Allow clicking not just on the menu item text, but on the whole "line" (icon, space next to the text etc.)
        self.menu.find(' > li').click(function (e) {
            if (e.target.nodeName != 'A') {
                // Some element for the menu item, other than the link itself, was clicked. Trigger a click on the link.
                var a = $(this).find('a');
                if (a.length > 0)
                    a[0].click(); // N.B. jQuery does not support triggering default anchor clicks, so use the DOM method
            }
        });        

        self.menu.show();
        // N.B. Menu must be visible before it can be positioned
        if (position)
            self.menu.position(position);

        if (self.fromExistingDOMElement && self.hasBeenOpened)
            self.menu.menu("refresh");
        else
            self.menu.menu();
        self.hasBeenOpened = true;
        self.menu.focus();

        currentlyOpenMenu = self;
    }        

    /**
     * Enables common keyboard shortcuts for the menu in addition to the default ones provided by jQueryUI
     * (Enter for "clicking" a menu item, Esc for closing the menu)
     */
    Menu.prototype.enableKeyboardShortcuts = function (callback) {
        var self = this;
            
        self.menu.keydown(function (e) {
            switch (e.which) {                    
                case 13: // enter
                    activeItem = self.menu.find('.ui-state-active a');
                    if (!activeItem.parents('li').hasClass('ui-state-disabled'))
                        activeItem[0].click(); // Use DOM method since jQuery click() does not support anchors with hrefs
                    e.stopPropagation();
                    break;
                case 27: // escape
                    self.close();
                    e.stopPropagation();
                    break;                
            }            
        });
    }

    return Menu;
});

