/*******************************************************************************
* Licensed Materials - Property of HCL
* (c) Copyright HCL Technologies Ltd. 2016, 2018.  All Rights Reserved.
*
* Note to U.S. Government Users Restricted Rights:
* Use, duplication or disclosure restricted by GSA ADP Schedule
* Contract with IBM Corp.
*******************************************************************************/

/**
 * Misc utilities
 *
 * Note: In this file we cannot use "modern" JavaScript since it will be loaded in Jazz webpages which have
 * ===== set compatibility mode to older versions of Internet Explorer where such JavaScript cannot be used!
 *
 * @author Mattias Mohlin
 */

define([], function () {
    function Utils() { }

    // Browser detection functions below. Only use when absolutely necessary.

    // Determines if the current browser is Firefox
    Utils.isFirefox = function () {
        return navigator.userAgent.toLowerCase().indexOf('firefox') > -1;
    }

    // Determines if the current browser is Chrome
    Utils.isChrome = function () {
        return /Chrome/.test(navigator.userAgent) && /Google Inc/.test(navigator.vendor);
    }

    // Escapes a regular expression string to make sure it can be parsed as a RegExp
    Utils.escapeRegExp = function (regexp) {
        return regexp.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&");
    }

    // Escapes HTML characters
    Utils.escapeHtml = function (text) {
        if (typeof text != "string")
            return "";

        var map = {
            '&': '&amp;',
            '<': '&lt;',
            '>': '&gt;',
            '"': '&quot;',
            "'": '&#039;'
        };

        return text.replace(/[&<>"']/g, function (m) { return map[m]; });
    }

    // Un-escapes HTML characters
    Utils.unescapeHtml = function (text) {
        if (typeof text != "string")
            return "";

        var map = {
            '&amp;': '&',
            '&lt;': '<',
            '&gt;': '>',
            '&quot;': '"',
            '&#039;': "'"
        };

        return text.replace(/(&quot;|&lt;|&gt;|&amp;|&#039;)/g, function (m) { return map[m]; });
    }

    // Extracts an error messages from the 'err' object (coming from a failed server request)
    Utils.getErrorMessage = function (err) {
        if (err.hasOwnProperty('responseJSON'))
            return err.responseJSON.message;
        else if (err.hasOwnProperty('responseText') && err.responseText)
            return err.responseText;
        else if (err.hasOwnProperty('statusText'))
            return err.statusText;
        else
            return ""; // Failed to extract a meaningful error message
    }

    // Reveals a DOM element by scrolling its closest scroll parent if necessary
    Utils.revealInScrollParent = function (element) {
        var scrollParent = $(element).scrollParent();
        scrollParent.animate({ scrollTop: scrollParent.scrollTop() + $(element).position().top - 20 /* offset to see the whole element */}, { duration: 'medium', easing: 'swing' });
    }

    // Returns the value of a query variable in a URL. If a URL is not provided, the current browser URL is used (decoded)
    Utils.getQueryVariable = function (variable, url) {
        var query;
        if (url) {
            var p = document.createElement('a');
            p.href = url;
            query = p.search.substring(1);
        }
        else {
            query = window.location.search.substring(1);
        }
        
        var vars = query.split("&");
        for (var i = 0; i < vars.length; i++) {
            var index = vars[i].indexOf('=');            
            if (vars[i].substring(0, index) == variable) {
                return decodeURIComponent(vars[i].substring(index + 1));
            }
        }
        return (false);
    }

    // Add a query parameter (called 'key') with the specified value to a URL. Returns the new URL.
    Utils.addQueryParameterToURL = function(url, key, value) {        
        var re = new RegExp("([?&])" + key + "=.*?(&|$)", "i");
        var separator = url.indexOf('?') !== -1 ? "&" : "?";
        if (url.match(re)) {
            return url.replace(re, '$1' + key + "=" + value + '$2');
        }
        else {
            return url + separator + key + "=" + value;
        }
    }

    // Generate a unique id (for example useful when creating DOM elements that need a unique id).
    Utils.getUniqueId = function (){
        return Math.round(new Date().getTime() + (Math.random() * 100));
    }

    // Append a suffix to 'base' and return a string that is not contained in 'str'.
    // If 'base' itself isn't contained in 'str' it is returned without appending anything.
    Utils.makeNonContainedString = function (str, base) {
        if (str.indexOf(base) == -1)
            return base; // Already non-contained

        return Utils.makeNonContainedString(str, base + 'X');
    }

    // Merges those properties in obj2 that do not exist in obj1 into obj1.    
    Utils.mergeProperties = function (obj1, obj2) {
        for (var p in obj2) {
            if (!obj1.hasOwnProperty(p))
                obj1[p] = obj2[p];
        }
    }

    // Copies the text to the clipboard
    Utils.copyTextToClipboard = function (text) {
        var textArea = document.createElement("textarea");
        textArea.style.position = 'fixed';
        textArea.style.top = 0;
        textArea.style.left = 0;
        textArea.style.width = '2em';
        textArea.style.height = '2em';
        textArea.style.padding = 0;
        textArea.style.border = 'none';
        textArea.style.outline = 'none';
        textArea.style.boxShadow = 'none';
        textArea.style.background = 'transparent';
        textArea.value = text;
        document.body.appendChild(textArea);
        textArea.select();

        try {
            var successful = document.execCommand('copy');
        } catch (err) {
            console.log('Unable to copy "' + text + '" to clipboard!');
        }

        document.body.removeChild(textArea);
    }

    // Returns the language in which the user prefers to view web content.
    // On Chrome and Firefox this is a setting in the browser, while on IE the language is taken
    // from the Windows control panel "Regional settings".
    Utils.getPreferredLanguage = function () {
        if (navigator.languages && navigator.languages.length)
            return navigator.languages[0]; // Chrome & Firefox

        return navigator.language; // IE
    }

    return Utils;
});

