/**
 * Analyze structure of TCJS file
 * @param {*} _analyzer 
 * @param {*} source
 */
function tcjsAnalyzer(_analyzer, source) {
    // Parsing source
    let _ast = esprima.parseScript(source, { range: true, comment: true });

    // Match declaration against 'let tc = TCF.define(TCF.CPP_TRANSFORM);'
    function matchTCFDefine(decl) {
        for (let x = 0; x < decl.declarations.length; ++x) {
            let v = decl.declarations[x];
            if (v.id.name == 'tc') {
                if (v.init.type == 'CallExpression') {
                    let expr = v.init.callee;
                    if (expr.type == 'MemberExpression' &&
                        expr.object.name == 'TCF' &&
                        expr.property &&
                        expr.property.name == 'define') {
                        return true;
                    }
                }
            }
            return false;
        }
        return false;
    }

    // Process expression statement
    function visitExpressionStatement(decl) {
        if (decl.expression == null || decl.expression == undefined) {
            return;
        }
        switch (decl.expression.type) {
            case 'AssignmentExpression':
                visitAssignment(decl.expression.left, decl.expression.right, decl);
                break;
            default:
                break;
        }
    }

    // return context and field if expr is MemberExpression
    function matchFieldAccess(expr) {
        try {
            let result = { context: null, field: null };
            if (expr.type == 'MemberExpression') {
                result.context = expr.object.name;
                result.field = expr.property.name;
            }
            return result;
        } catch (error) {
            return { context: null, field: null };
        }
    }
    
    // Process assignment 
    function visitAssignment(left, right, decl) {
        let obj = matchFieldAccess(left);
        if (obj.context == 'tc') {
            // Adding location of assignment        
            _analyzer.addProperty(obj.field, decl.range[0], decl.range[1]);
        }
    }

    // Check if we have any declarations at all
    let analyze = true;
    if (_ast == null || _ast.body == null || _ast.body == undefined) {
        analyze = false;
    }

    if (analyze) {
         let hasTCFDefine = false;
        let bodyStart = null;
        let bodyEnd = null;

        for (let i = 0; i < _ast.body.length; ++i) {
            let decl = _ast.body[i];
            switch (decl.type) {
                case 'VariableDeclaration':
                    {
                        let isTCDefine = matchTCFDefine(decl);
                        if (isTCDefine && hasTCFDefine) {
                            _analyzer.setAdvancedSyntax();
                        } else if (!hasTCFDefine) {
                            hasTCFDefine = true;
                            bodyStart = decl.range;
                        }
                    }
                    break;
                case 'ExpressionStatement':
                    if (hasTCFDefine) {
                        visitExpressionStatement(decl);
                        bodyEnd = decl.range;
                    }
                    break;
                default:
                    break;
            }
        }
    }
}

/**
 * Build graph from list of TCJS files
 * @param {*} models 
 * @param {*} graph 
 */
function tcjsGraphBuilder(model, graph) {

    // Produce AST model from source code
    function parse(source) {
        try {
            let ast = esprima.parseScript(source);
            return ast;
        } catch (e) {
            return null;
        }
    }

    // Extract strings from array expression
    function parseRefs(expr, type) {
        if (expr == null) {
            return;
        }
        if (expr.type == 'ArrayExpression') {
            for (let i = 0; i < expr.elements.length; ++i) {
                let item = expr.elements[i];
                if (item.type == 'Literal') {
                    graph.addRef(type, item.value);
                }
            }
        }
    }

    // Get TC type
    function parseType(expr) {
        if (expr.type == 'MemberExpression') {
            graph.setType(expr.property.name);
        }
    }

    // return context and field if expr is MemberExpression
    function matchFieldAccess(expr) {
        try {
            let result = { context: null, field: null };
            if (expr.type == 'MemberExpression') {
                result.context = expr.object.name;
                result.field = expr.property.name;
            }
            return result;
        } catch (error) {
            return { context: null, field: null };
        }
    }
    

    // Return object with parents and prerequisites references
    function buildGraph(ast) {
        let result = {
            parents: [],
            prerequisites: []
        };

        for (let i = 0; i < ast.body.length; ++i) {
            let decl = ast.body[i];
            switch (decl.type) {
                case 'ExpressionStatement':
                    switch (decl.expression.type) {
                        case 'AssignmentExpression':
                            {
                                let left = decl.expression.left;
                                let right = decl.expression.right;
                                let context = matchFieldAccess(left);
                                if (context.field == 'parents' || context.field == 'prerequisites') {
                                    parseRefs(right, context.field);
                                } else if (context.field == 'type') {
                                    parseType(right);
                                }
                            }
                            break;
                        default:
                            break;
                    }
                    break;
                default:
                    continue;
            }
        }
    }
    let ast = parse(model);
    if (ast == null) {
        return;
    }
    buildGraph(ast);
}