/**
 * Build TCFile model in Java
 * 
 * @param {*} _logger 
 * @param {*} _valueBuilder 
 * @param {*} _valueChecker
 * @param {*} script
 * @param {*} _model 
 */
function tcjsBuilder(_logger, _valueBuilder, _valueChecker, script, _model) {
    let wereErrrors = false;

    // Parsing source
    let _ast = esprima.parseScript(script, { range: true, comment: true });

    // issue error at specific position
    function error(range, msg) {
        wereErrrors = true;
        _logger.error(range[0], range[1], msg);
    }

    function advancedSyntax(range, msg) {
        wereErrrors = true;
        _model.setAdvancedSyntax();
        if (!msg)
            msg = "JavaScript statement beyond simple assignment";
        else
            msg = "JavaScript statement beyond simple assignment: " + msg;
        _logger.error(range[0], range[1], msg);
    }

    function pos(i, l) {
        return [i, l];
    }

    function convertValue(property, expr) {
        let value = null;
        let rawValue = script.slice(expr.range[0], expr.range[1]);
        if (expr.type == 'ArrayExpression') {
            value = _valueBuilder.makeStringList();
            for (let i = 0; i < expr.elements.length; ++i) {
                let item = expr.elements[i];
				if (item == null)
					continue;
                if (item.type == 'Literal') {
                    value.add(item.value);
                } else {
                    return _valueBuilder.makeRawValue(rawValue);
                }
            }
        } else if (expr.type == 'MemberExpression') {
            value = expr.object.name + '.' + expr.property.name;
            if (!_valueChecker.isValidValue(property, value)) {
                return _valueBuilder.makeRawValue(rawValue);
            }
        } else if (expr.type == 'Literal') {
            value = expr.value;
            _model.setQuote(property, 1);
        } else if (expr.type == 'TemplateLiteral') {
            _model.setQuote(property, 2);
            let begin = expr.range[0] + 1;
            let end = expr.range[1] - 1;
            if (begin <= end) {
                value = script.slice(begin, end).trim();
            } else {
                value = '';
            }
        } else if (expr.type) {
            if (_valueChecker.isValidValue(property, expr)) {
                value = script.slice(expr.range[0], expr.range[1]);
            } else {
                return _valueBuilder.makeRawValue(rawValue);
            }
        }
        return value;
    }

    // Add property value to TC model
    function addProperty(prop, expr) {        
        if (prop == 'threads') {
            try {
                _model.addProperty(prop, getThreadsProperty(expr));
            } catch (e) {
                advancedSyntax(e.expr.range, e.message);
            }
        } else if (_model.isKnownProperty(prop)) {
            let value = convertValue(prop, expr);
            if (value != null && !wereErrrors) {
                _model.addProperty(prop, value);
            }
        } else {
            let rawValue = script.slice(expr.range[0], expr.range[1]);
            _model.addProperty(prop, _valueBuilder.makeRawValue(rawValue));
        }
    }

    function threadsEvalError(msg, expr) {
        throw { message: msg, expr: expr };
    }

    function threadsCheckType(expr, types) {
        types = [].slice.call(arguments);
        types.shift();
        if (!types.some(function(t) { return t === expr.type; }))
            threadsEvalError('Unexpected type in threads expression. Expected: ' + types.join(', ') + ', actual: ' + expr.type, expr);
        return expr.type;
    }

    function getThreadObject(expr) {
        return expr.properties.reduce(function(e, p) {
            threadsCheckType(p.key, 'Identifier');
            let prop = p.key.name;

            let val;
            if (prop === 'logical') {
                // logical 
                threadsCheckType(p.value, 'ArrayExpression');
                val = p.value.elements.map(function(elt) {
                    threadsCheckType(elt, 'Literal');
                    return String(elt.value);
                });
            } else {
                threadsCheckType(p.value, 'Literal');
                val = String(p.value.value);
            }

            let v = new Object(e);
            v[prop] = val;
            return v;
        }, ({}));
    }

    function getThreadsProperty(expr) {
        threadsCheckType(expr, 'ArrayExpression');
        let elementType = null;
        return expr.elements.map(function(element) {
            if (elementType) {
                threadsCheckType(element, elementType);
            } else {
                elementType = threadsCheckType(element, 'Literal', 'ObjectExpression');
            }
            return elementType === 'Literal' ? String(element.value) : getThreadObject(element);
        });
    }

    // Match declaration against 'let tc = TCF.define(TCF.CPP_TRANSFORM);'
    function matchTCFDefine(decl) {
        for (let x = 0; x < decl.declarations.length; ++x) {
            let v = decl.declarations[x];
            if (v.id.name == 'tc') {
                if (v.init.type == 'CallExpression') {
                    let expr = v.init.callee;
                    if (expr.type == 'MemberExpression' &&
                        expr.object.name == 'TCF' &&
                        expr.property &&
                        expr.property.name == 'define') {
                        return true;
                    }
                }
            }
            return false;
        }
        return false;
    }

    // Process expression statement
    function visitExpressionStatement(decl) {
        if (decl.expression == null || decl.expression == undefined) {
            return;
        }
        switch (decl.expression.type) {
            case 'AssignmentExpression':
                _model.setCurrentRange(decl.range[0], decl.range[1]);
                visitAssignment(decl.expression.left, decl.expression.right);
                break;
            default:
                advancedSyntax(decl.range);
        }
    }

    // Process assignment 
    function visitAssignment(left, right) {
        let obj = matchFieldAccess(left);
        if (obj.context == 'tc') {
            addProperty(obj.field, right);
        } else {
            advancedSyntax(left.range);
        }
    }

    // return context and field if expr is MemberExpression
    function matchFieldAccess(expr) {
        try {
            let result = { context: null, field: null };
            if (expr.type == 'MemberExpression') {
                result.context = expr.object.name;
                result.field = expr.property.name;
            }
            return result;
        } catch (error) {
            return { context: null, field: null };
        }
    }

    // Check if we have any declarations at all
    if (_ast == null || _ast.body == null || _ast.body == undefined) {
        error(pos(0, 0), "The file is empty");
    }

    let hasTCFDefine = false;
    let bodyStart = null;
    let bodyEnd = null;

    for (let i = 0; i < _ast.body.length; ++i) {
        let decl = _ast.body[i];
        switch (decl.type) {
            case 'VariableDeclaration':
                {
                    if (matchTCFDefine(decl)) {
                        if (hasTCFDefine) {
                            advancedSyntax(decl.range);
                        } else {
                            hasTCFDefine = true;
                            bodyStart = decl.range;
                            _model.setCurrentRange(decl.range[0], decl.range[1]);
                        }
                    } else {
                        advancedSyntax(decl.range);
                    }
                }
                break;
            case 'ExpressionStatement':
                if (hasTCFDefine) {
                    visitExpressionStatement(decl);
                    bodyEnd = decl.range;
                } else {
					advancedSyntax(decl.range);
				}
                break;
            default:
                advancedSyntax(decl.range);
                break;
        }
    }
    if (bodyStart) {
        _model.setPreface(script.slice(0, bodyStart[0]));
    }
    if (bodyEnd) {
        _model.setEnding(script.slice(bodyEnd[1]));
    }
}