/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTActor_h__
#define __RTActor_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTController_h__
#include <RTController.h>
#endif

#ifndef __RTVersionId_h__
#include <RTVersionId.h>
#endif

struct RTActorClass;
class  RTActorProbe;
class  RTActorRef;
struct RTActor_class;
struct RTBindingEnd;
struct RTComponentDescriptor;
class  RTDebugger;
union  RTImportList;
class  RTMessage;
struct RTObject_class;
class  RTProtocol;
class  RTRelayPort;

/** An instance of this class represents a capsule instance. All classes that are generated for
 capsules inherit from this class.
 @see RTActorId is a wrapper object around an RTActor pointer and hence also represents a capsule
 instance
 */
class RTActor
{
    friend class RTActorRef;
    friend class RTController;
    friend class RTDebugger;
    friend class RTMessage;
    friend class RTProtocol;

private: // instance data
    RTController      * rts;
    RTActorRef        * homeReference;
    int                 homeIndex;
    RTRelayPort       * relays;
#if DEFER_IN_ACTOR
    RTMessageQ          deferredQ;
#endif
#if SUSPEND
    RTMessage         * suspendMsgs; // stack of messages which arrived while the actor was suspended
#endif

    RTImportList      * importList;

    struct State
    {
    	RTStateId          current;
#if RTStateId_MaxSize > 1
    	unsigned char      idShift; // log2( log256( getNumStates() ) )
#endif
    	union
		{
    		unsigned char  s1[ 1 ];
#if RTStateId_MaxSize > 1
    		unsigned short s2[ 1 ];
#endif
#if RTStateId_MaxSize > 2
    		unsigned int   s4[ 1 ];
#endif
		} history;
    };

    State             * state;

public: // instance data

    /** The message which is currently being processed by the capsule instance. */
    const RTMessage   * msg;

#if OBSERVABLE
    /** A probe attached to the capsule instance for monitoring it. */
    RTActorProbe      * probe;
#endif

protected: // functions

    /** @deprecated This function does nothing. */
    RTS_INLINE void         saveHistory( void );

    /** Force the state machine of the capsule instance into a particular state.
     @param[in] s The numeric id of a state to be set as the active state. Make sure to pass an
     id of a valid state, defined in the state machine of the capsule instance. If the state machine
     is forced into an invalid state, unexpectedState() will be called.
     */
    RTS_INLINE void         setCurrentState( int s);

    /** This function is called when a message is dispatched to the capsule instance, and when
     there is no enabled transition in the current state that can be triggered to handle the
     received message. The default implementation prints a message to stderr. You may override
     the function to provide a different behavior.
     */
    virtual void            unexpectedMessage( void );

    /** This function is called if the TargetRTS detects that the state machine of the capsule
     instance in in an invalid state (i.e. the current state does not match any of the states
     defined in the capsule state machine). This is an abnormal situation which will cause the
     application to terminate.
     */
    void                    unexpectedState( void );

#if ! OBSERVABLE
    RTS_INLINE
#endif
	/** This function is called at the beginning of every generated chain function that corresponds
	 to a transition in the model. The parameters identifies the start of the transition chain.
	 Note that the first argument is implied by the second, but skipping it would require generation
	 of a mapping table which, on some architectures, could be larger than the code which passes the
	 argument.
	 @param[in] origin Identifies the start point of the chain. Numbers from 1 to getNumStates(),
	 inclusive, refer to states, larger numbers refer to choice-points.
	 @param[in] trans The name of the transition.
	 */
    void                    rtgChainBegin( int origin, const char * trans );

#if ! OBSERVABLE
    RTS_INLINE
#endif
    /** This function is called just before a user-defined transition effect executes. */
    void                    rtgTransitionBegin( void );

#if ! OBSERVABLE
    RTS_INLINE
#endif
    /** This function is called just after a user-defined transition effect has executed. */
    void                    rtgTransitionEnd( void );

    void                    enterState( int );
    virtual void            enterStateV( void );

    /** This function is called when the state machine of the capsule instance enters a new state.
     The call is made immediately after the state has become active, before the entry action of
     that state executes. The default implementation of this function does nothing. The function
     can be overridden on a capsule that wants to perform some common actions whenever a new state
     in its state machine becomes active. To override this function, simply add a new operation to
     the capsule in the model with the same name and prototype.
     */
    virtual void            rtgStateEntry( void );

    void                    exitState( const RTStateId parent[] );
    virtual void            exitStateV( void );
    void                    processHistory( void );
    void                    exitToChainState( int, const RTStateId parent[] );

    void                    initializeAll( void );

    int                     removeImport( RTActorRef * ,RTController * initiator );
    void                    removeAllImports( RTController * initiator );

    int                     _followIn( RTBindingEnd & end,
				       	   	   	   	   int portId,
									   int repIndex,
									   RTActorRef * importer );

    int                     _followOut( RTBindingEnd & end,
										int portId,
										int repIndex );

    void                    _unbind( RTActorRef * importer  ,RTController* initiator);
    virtual void            _predestroy( void );

public: // functions

    /** Constructor. Never create a capsule instance directly using the new operator. Always
     create a capsule instance using Frame::Base::incarnate(), or automatically at start-up by means
     of fixed capsule parts.
     */
	RTActor( RTController *, RTActorRef * );

    virtual ~RTActor( void );

    /** Get the index of this capsule instance in its container capsule part (the capsule part
     where it was incarnated).
     @return A zero-based index.
     */
    RTS_INLINE int          getIndex( void ) const;

    /** Get the index of this capsule instance in its container capsule part (the capsule part
	 where it was incarnated).
	 @return A 1-based index.
	 */
    RTS_INLINE int          index( void ) const;

    /** Get the numeric id for the capsule part that contains this capsule instance.
     @return A numeric id for the container capsule part that is unique within the capsule
     */
    RTS_INLINE int          getId( void ) const;

    /** Get the name of the capsule part that contains this capsule instance.
     @return The name of the container capsule part. For the top capsule instance the string "Top"
     is returned.
     */
    const char            * getName( void ) const;

    /** Get the capsule part that contains this capsule instance (the capsule part where it was
     * incarnated).
	 @return The container capsule part.
	 */
    RTS_INLINE RTActorRef * getReference( void ) const;

    /** Get the message which is currently being processed by the capsule instance.
     @return The currently processed message.
     */
    RTS_INLINE const RTMessage * getMsg( void );

    /** Make a copy of a message and send it to this capsule instance. This function can for example
     be used in order to defer the handling of a reply message, or in general where a message cannot
     be completely handled in a single transition.
     @param[in] msg A message (typically the one currently being processed by the capsule instance).
     @return 1 in case the message was successfully copied and sent, 0 otherwise.
     */
    RTS_INLINE int          sendCopyToMe ( const RTMessage* msg );

    /** Defines the state machine behavior for this capsule instance. This function is implemented
     in each concrete capsule subclass.
     @param[in] signal The event of the currently processed message.
     @param[in] port The port on which the currently processed message arrived.
     */
    virtual void            rtsBehavior( int signal, int port ) = 0;

#if LOG_MESSAGE
    /** This function is called just before a received message is processed by the capsule instance.
     The default implementation prints a message to the log (depending on the current debug level).
     Override this function to perform any other general action before a message is processed.
     */
    virtual void            logMsg( void );
#endif

    /** Get the controller which runs this capsule instance.
     @return The controller object which runs this capsule instance.
     */
    RTS_INLINE RTController * context( void ) const;

    /** Determines if this capsule instance is currently active, i.e. is processing a message.
     @return 1 if the capsule instance is active, 0 otherwise.
     */
    RTS_INLINE int          isActive( void ) const;

    virtual int             designLevel( void ) const;

    /** Get the name of the capsule as defined in the model.
     @return Name of the capsule instance type.
     */
    RTS_INLINE const char * getTypeName( void ) const;

    RTS_INLINE RTVersionId  getVersion( void ) const;

    /** Determines if the type of this capsule instance is a particular capsule type.
     @param[in] cap A capsule type.
     @return 1 if the type of this capsule instance equals the capsule type, 0 otherwise.
     */
    RTS_INLINE int          isSameType( const RTActorClass * cap) const;

    /** Performs a lexical comparison between the name of the capsule with a specified string.
     @param[in] str A string to compare the capsule name with.
     @return 0 if the capsule type name equals the string. A negative or positive value is returned
     depending on the result of comparing the first character that is different.
     */
    int                     isType( const char * str) const;

#if SUSPEND
    void                    suspend( void );
    void                    resume( void );
    RTS_INLINE int          isSuspended( void ) const;
#endif

    /** Get information (meta data) about the capsule, such as its name, ports, states etc. This
     information is static and comes from the definition of the capsule in the model.
     @return Information about the capsule.
     */
    virtual const RTActor_class  * getActorData( void ) const = 0;

    /** Get the number of states in the capsule state machine.
     @return Number of state machine states.
     */
    int                     getNumStates( void ) const;

    /** Get the name of one of the states in the capsule state machine.
     @param[in] s The numeric identifier of a state.
     @return The name of the specified state. An error string is returned in case of error (for
     example if an invalid state id is specified).
     */
    const char            * getStateString( int s) const;

    /** Get the name of the currently active state in the capsule state machine.
     @return The name of the currently active state.
     */
    const char            * getCurrentStateString( void ) const;

    /** Get the numeric id of the currently active state in the capsule state machine.
	 @return The id of the currently active state.
	 */
    RTS_INLINE int          getCurrentState( void ) const;

    /** Get the numeric id of the previously active state in the capsule state machine.
	 @return The id of the state that was most previously active.
	 */
    RTS_INLINE int          getHistory( int ) const;

    int                     _numComponents( void ) const;
    RTActorRef            * _component( int compId );

    void                    _preinitialize( int _homeIndex,
					    					const RTComponentDescriptor * );
    int                     initMe( const void           * data,
				    				const RTObject_class * type,
									RTController         * sender );

#if RTIMPORT_ISREFERENCEDBY_CHECK
    int                     isReferencedBy( const RTActor * ) const;
#endif

    /** Get the default capsule class for a capsule part contained in the capsule instance.
     This function is used by the Frame service when incarnating capsules without specifying
     the capsule type. The default implementation just returns the capsule that types the capsule
     part. You may override the function to use a different capsule as the default capsule class.
     @param[in] component A capsule part of the capsule instance.
     @param[in] repIndex The index within the capsule part where the Frame service will insert the
     capsule instance to be created. It is unused in the default implementation.
     @return The default capsule class for the capsule part.
     */
    virtual const RTActorClass * defaultClass( RTActorRef & component,
					       	   	   	   	   	   int          repIndex );

    /** Get the default controller for a capsule part contained in the capsule instance.
     This function is used by the Frame service when incarnating capsules without specifying
     the controller to run the created capsule instance. The default implementation just returns
     the controller of this capsule instance. You may override the function to use a different
     controller as the default.
     @param[in] component A capsule part of the capsule instance.
     @param[in] repIndex The index within the capsule part where the Frame service will insert the
     capsule instance to be created. It is unused in the default implementation.
     @return The default controller for the capsule part.
     */
    virtual RTController  * defaultController( RTActorRef & component,
					       	   	   	   	   	   int          repIndex );

    void                    _incarnateAllFixed( void );

    // The functions below are generated by the toolset.
    virtual int             _followOutV( RTBindingEnd & end,
										 int            compId,
										 int            portId,
										 int            portIndex );

    // This function is simply inherited unless the actor has additional
    // (interface) ports or has excluded ports of its superclass.
    virtual int             _followInV( RTBindingEnd & end,
										int            portId,
										int            repIndex );

protected:
    /** Get the most recent error that occurred when using the TargetRTS. Call this function
     immediately after a function has returned with an error code to find out more information about
     the error that occurred.
     @return An RTController::Error value describing the error that occurred most recently when
     using the TargetRTS.
     */
    RTS_INLINE RTController::Error getError( void ) const;

private:
    // unavailable methods
    RTActor( const RTActor & );
    RTActor & operator=( const RTActor & );
}; //lint !e1712

#if RTS_INLINES
#include <RTActor.inl>
#endif

#endif // __RTActor_h__
