/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTActorRef_h__
#define __RTActorRef_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTConfig_h__
#include <RTConfig.h>
#endif

class  RTActor;
struct RTActorClass;
class  RTActorFactoryInterface;
class  RTActorFactory;
class  RTActorId;
class  RTActorRefProbe;
struct RTBindingEnd;
struct RTComponentDescriptor;
class  RTController;
class  RTDataObject;
union  RTImportList;
struct RTObject_class;

/** Represents a capsule part. For each capsule part in the composite structure of a capsule a
 variable of this type is added to the RTActor subclass that is generated for that capsule. A
 capsule part can at run-time contain capsule instances (zero to many).
 */
class RTActorRef
{
private:
    const RTComponentDescriptor   * _info;
    void                          * _contents;

#if USE_THREADS
    friend class RTActor;

    static void                 lock( void );
    static void                 unlock( void );
#endif

    RTActor                  ** access( int ) const;
    int                         reserve( int & );
    void                        unreserve( int );
    void                        add( RTActor *, int );
    void                        remove( int );

public:
#if OBSERVABLE
    RTActorRefProbe           * probe;
#endif

    /** Get the capsule to which this capsule part belongs.
	 @return The capsule that contains this capsule part in its composite structure.
	 */
    RTS_INLINE RTActor        * getOwner( void ) const;

    /** Get the maximum number of capsule instances that can fit in the capsule part (as defined
     by its upper multiplicity in the model).
	 @return The size of the capsule part.
	 */
    RTS_INLINE int              size( void ) const;

    /** Get the id of the capsule part. Each capsule part has a numeric id that is unique within
     the composite structure of the owner capsule.
	 @return The id of the capsule part.
	 */
    RTS_INLINE int              getId( void ) const;

    /** Get the name of the capsule part as defined in the model.
	 @return The name of the capsule part.
	 */
    RTS_INLINE const char     * getName( void ) const;

    /** Get information (meta data) about the capsule part, such as its name, size etc. This
	 information is static and comes from the definition of the capsule part in the model.
	 @return Information about the capsule part.
	 */
    RTS_INLINE const RTComponentDescriptor * getInfo( void ) const;

    /** Get the controller which runs the capsule that owns this capsule part.
	 @return The controller object which runs the owner capsule instance.
	 */
    RTS_INLINE RTController   * context( void );

    /** Incarnate capsule instances into a fixed capsule part. You should not call this function
     explicitly - it will be implicitly called when the container capsule instance is created.
     That is, immediately after a capsule instance has been created all its fixed capsule parts
     will already contain all capsule instances they can contain.
	 */
    void                        incarnateAll( void );

    /** Initialize capsule instances that have been incarnated in a fixed capsule part. You should
     not call this function explicitly - it will be implicitly called when the container capsule
     instance is created. That is, immediately after a capsule instance has been created all its
     fixed capsule parts will already contain all capsule instances they can contain, and they
     will already be initialized.
	 */
    void                        initializeAll( void );

    /** Incarnates a capsule and inserts the capsule instance in this capsule part (which must be
     an optional capsule part). Consider using RTFrame::incarnate() instead of calling this function
     directly.
	 @param[in] cap The capsule to incarnate.
	 @param[in] data Initialization data to pass to the created capsule instance. It can be
	 accessed from the initial transition of its state machine.
	 @param[in] type Type descriptor for the data object.
	 @param[in] controller The controller which should run the created capsule instance. If it
	 is NULL, the same controller that runs the container capsule instance will be used.
	 @param[in] index The index within a replicated capsule part where to insert the created
	 capsule instance. Use -1 to insert the instance at the first free slot.
	 @param[in] factory If provided this capsule factory will be used for creating the capsule instance.
	 It can be useful if the capsule has template parameters or a custom capsule constructor should be used.
	 @return The created capsule instance.
	 */
    RTActor                   * incarnate( const RTActorClass   & cap,
										   const void           * data,
										   const RTObject_class * type,
										   RTController         * controller,
										   int                    index,
										   RTActorFactory		* factory = nullptr);

    /** Destroy all capsule instances contained in the capsule part (which must be
     an optional capsule part). Consider using RTFrame::destroy() instead of calling this function
     directly.
     @return The number of destroyed capsule instances. 0 is returned in case of errors.
	 */
    int                         destroy( void );

    /** Destroy a capsule instance contained in the capsule part (which must be an optional capsule
     part). Consider using RTFrame::destroy() instead of calling this function directly.
	 @param[in] i The capsule instance to destroy.
	 @return 1 if the capsule instance was successfully destroyed, otherwise 0.
	 */
    int                         destroy( RTActor * i );

    /** Import a capsule instance into this capsule part (which must be a plug-in capsule part typed
     by a capsule that is compatible with the type of the imported capsule instance).
     Consider using RTFrame::import() instead of calling this function directly.
	 @param[in] i Capsule instance to import. It must be located in an optional capsule part
	 within the same composite structure as this capsule part.
	 @param[in] index The index within this capsule part where to insert the imported
	 capsule instance. Use -1 to insert the instance at the first free slot.
	 @return 1 if the capsule instance was successfully imported, 0 otherwise.
	 */
    int                         import( RTActor * i, int index);

    /** Remove all capsule instances from a plug-in capsule part.
	 */
    void                        deport( void );

    /** Remove a capsule instance from a plug-in capsule part. Consider using RTFrame::deport()
     instead of calling this function directly.
	 @param[in] i The capsule instance to remove (deport).
	 @return 1 if the capsule instance was successfully deported, 0 otherwise.
	 */
    int                         deport( RTActor * i);

    /** A helper function for internal use only. Call RTFrame::deport() instead.
     */
    int                         deport( RTActor *, RTImportList *, RTController * initiator );

    /** Determines if this capsule part is fixed, i.e. if it always holds a fixed number of capsule
     instances.
	 @return 1 if the capsule part is fixed, 0 otherwise.
	 */
    RTS_INLINE int              isFixed( void ) const;

    /** Determines if this capsule part is optional, i.e. if it may contain a variable number of
     capsule instances (including none).
	 @return 1 if the capsule part is optional, 0 otherwise.
	 */
    RTS_INLINE int              isOptional( void ) const;

    /** Determines if this is a plug-in capsule part into which capsule instances can be imported.
	 @return 1 if the capsule part is a plug-in capsule part, 0 otherwise.
	 */
    RTS_INLINE int              isImported( void ) const;

    /** Get the capsule instance contained in this capsule part at a certain index (0-based).
	 @param[in] index Zero-based index.
	 @return The capsule instance at the specified index, or NULL if none.
	 */
    RTS_INLINE RTActor        * operator[]( int index) const;

    /** Get the capsule instance contained in this capsule part at a certain index (1-based).
	 @param[in] index One-based index.
	 @return The capsule instance at the specified index, or NULL if none.
	 */
    RTS_INLINE RTActor        * at( int index) const;

    /** Return the index of a capsule instance within this capsule part.
	 @param[in] i Capsule instance
	 @return The index (0-based) of the capsule instance within this capsule part. -1 is returned
	 in case of errors (e.g. when the capsule instance is not contained in the capsule part).
	 */
    int                         indexOf( const RTActor * i) const;

    RTS_INLINE                  RTActorRef( void );
								~RTActorRef( void );

    RTS_INLINE int              operator==( const RTActorId & ) const;
    RTS_INLINE int              operator!=( const RTActorId & ) const;

    void                        init( const RTComponentDescriptor * info );

    int                         _followIn( RTBindingEnd & end,
					   	   	   	   	   	   int            portId,
										   int            portIndex );
    int                         _followOut( RTBindingEnd & end,
					    					RTActor      * member,
											int            portId,
											int            portIndex );

private:
    int                         bind_ports( RTActor    * member,
					    					int          index,
											RTActorRef * importer );

    RTActor                   * incarnate2( const RTActorClass *,
					    					RTController       *,
											int 						 index,
											RTActorFactoryInterface	   * factory = nullptr );

    // unavailable methods
    RTActorRef( const RTActorRef & );
    RTActorRef & operator=( const RTActorRef & );
};

#if RTS_INLINES
#include <RTActorRef.inl>
#endif

#endif // __RTActorRef_h__
