/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTAsciiEncoding_h__
#define __RTAsciiEncoding_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTEncoding_h__
#include <RTEncoding.h>
#endif

#if OBJECT_ENCODE

class RTOBuffer;

#if OBJECT_DECODE && ! defined( __RTAsciiDecoding_h__ )
#include <RTAsciiDecoding.h>
#endif

/** The default encoding implementation that is used for all default type descriptors.
 Types are encoded as ASCII strings by printing the type name followed by the typed value.
 Examples: int 8   double 2.7   MyClass{a 8, b false}
 */
class RTAsciiEncoding : public RTEncoding
{
public:
	/** Construct an ASCII encoder object.
	 @param buffer The output buffer to which the encoded string should be written.
	 */
    explicit    RTAsciiEncoding( RTOBuffer * buffer);
    virtual     ~RTAsciiEncoding( void );

    // All the get/put functions return an int which is non-zero
    // if the operation was successful. Otherwise, the following
    // function may be called to discover the reason for the
    // failure.
    int         getError( void );

    /** Flush the output buffer. Everything this encoder has written so far will appear in the
     output.
     */
    virtual int flush( void ) override;

    /** Encode a data object and write the encoding to the output buffer.
     @param data Data object to encode.
     @param type Type descriptor describing the type of the data object.
     @return 1 if successfully encoded and written to the output buffer, 0 otherwise
     */
    virtual int put( const void * data, const RTObject_class * type) override;

    // Functions which deal with the basic data types.
    virtual int put_address ( const void   * ) override;
    virtual int put_bool    ( bool           ) override;
    virtual int put_char    ( char           ) override;
#if RTUseFloatingPoint
    virtual int put_double  ( double         ) override;
    virtual int put_float   ( float          ) override;
#endif
    virtual int put_int     ( int            ) override;
    virtual int put_long    ( long           ) override;
    virtual int put_short   ( short          ) override;
    virtual int put_uchar   ( unsigned char  ) override;
    virtual int put_unsigned( unsigned       ) override;
    virtual int put_ulong   ( unsigned long  ) override;
    virtual int put_ushort  ( unsigned short ) override;

    virtual int put_unsignedint( unsigned int) override;
    virtual int put_unsignedlong( unsigned long) override;
	virtual int put_unsignedshort( unsigned short) override;
	virtual int put_unsignedchar( unsigned char) override;
    virtual int put_string( const char * ) override;

    virtual int put_opaque( int, const char * ) override;

    virtual int put_enum( int, int, const RTFieldDescriptor * ) override;

    virtual int put_array( const void *, int, const RTObject_class * ) override;

    virtual int put_indirect( const void *, const RTObject_class * ) override;

    virtual int put_struct( const void *, const RTObject_class * ) override;

    /** Write a string verbatimely to the output buffer of the encoder.
     @param str String to write to the output buffer. It must not contain any unprintable
     characters.
     @return 1 if successful, 0 otherwise
     */
    virtual int write_string( const char* str) override;

protected:
    enum Spacing { SpaceNone, SpaceAround };

    RTOBuffer     * output;
    Spacing         before;

    virtual int     putType( const RTObject_class * );
    int             putName( const char * );

    virtual int     putFields( const void *, const RTObject_class *, int & first );

    int             putFieldListStart( void );
    int             putFieldSeparator( void );
    int             putFieldListFinish( void );

    int             putCode( const char *, int, Spacing );

private:
    // unavailable methods
    RTAsciiEncoding( const RTAsciiEncoding & );
    RTAsciiEncoding & operator=( const RTAsciiEncoding & );
}; //lint !e1712

#endif // OBJECT_ENCODE

#endif // __RTAsciiEncoding_h__
