/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTController_h__
#define __RTController_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTJob_h__
#include <RTJob.h>
#endif

#ifndef __RTMessageQ_h__
#include <RTMessageQ.h>
#endif

#ifndef __RTPriority_h__
#include <RTPriority.h>
#endif

#ifndef __RTResourceMgr_h__
#include <RTResourceMgr.h>
#endif

#if RTS_COUNT && ! defined( __RTCounts_h__ )
#include <RTCounts.h>
#endif

class  RTActor;
struct RTActorClass;
struct RTComponentDescriptor;
class  RTDaemonInfo;
class  RTDebugger;
class  RTEventInfo;
class  RTSyncObject;
class  RTTimerNode;

/** A controller manages a group of capsule instances that all run in the same physical thread
 (i.e. there is exactly one controller object for each physical thread in the RT application).
 The controller implements the mechanism for dispatching messages to capsule instances. RTController
 is an abstract class, and its concrete subclasses have different implementations of how messages
 are dispatched. By default the RTPeerController subclass is used as the controller for capsule
 instances that run in the same physical thread.
 */
class RTController : public RTJob
{
private:
    friend class RTMessage;
    friend class RTProtocol;

    const char        * _name;

#if OTRTSDEBUG != DEBUG_NONE
    friend class RTDebugger;

    RTDebugger        * _debugger;
    RTController      * _prev;
    RTController      * _next;

    enum
    {
		detached,       // execution is independent of the debugger
		execSystem,     // executing in system code (between chains)
		waitBegin,      // awaiting permission to begin executing a chain
		waitDone,       // executing in system code (in a chain)
		execUser,       // executing in user code (in a chain)
		waitContinue    // awaiting permission to continue executing a chain
    } _state;
#endif // OTRTSDEBUG != DEBUG_NONE

public:
	RTController( RTDebugger * debug,
			      const char * name );
    virtual             ~RTController( void );

    enum Primitive
    {
		CommDefer,
		CommDeliver,
		CommInvoke,
		CommPurge,
		CommRecall,
		CommRecallAll,
		CommReply,
		CommSend,
		ExceptionRaise,
		FrameDeport,
		FrameDestroy,
		FrameImport,
		FrameIncarnate,
		LayerBind,
		LayerRegisterSAP,
		LayerDeregisterSAP,
		LayerRegisterSPP,
		LayerDeregisterSPP,
		LayerResize,
		LayerResizeSAP,
		LayerResizeSPP,
		TimerCancel,
		TimerInform,

		SystemService
    };

    enum Error
    {
		ok,

		// code             triggered by
		// ====             ============

		alreadyDeferred, // CommDefer
		badActor,        // FrameDeport FrameDestroy FrameImport
		badClass,        // CommDeliver CommInvoke CommReply CommSend FrameImport FrameIncarnate
		badId,           // TimerCancel
		badIndex,        // FrameIncarnate CommInvoke CommSend
		badInterface,    // FrameImport FrameIncarnate
		badOperation,    // LayerDeregister LayerRegister LayerResize{,SAP,SPP}
		badRef,          // FrameImport FrameIncarnate
		badSignal,       // CommInvoke CommReply CommSend
		badValue,        // LayerResize{,SAP,SPP}
		deferInvoke,     // CommDefer CommInvoke
		dereg,           // LayerDeregister
		imported,        // FrameDestroy FrameIncarnate
		noConnect,       // CommInvoke CommReply TimerInform
		noFreePorts,     // FrameImport
		noMem,           // all
		noReply,         // CommInvoke
		notImported,     // FrameDeport FrameImport
		notOptional,     // FrameDestroy
		notOwner,        // FrameDeport FrameDestroy FrameImport FrameIncarnate
		prio,            // CommReply CommSend TimerInform
		protocolSignal,  // FrameImport FrameIncarnate LayerBind
		protocolType,    // FrameImport FrameIncarnate LayerBind
		recursiveInvoke, // CommInvoke
		refFull,         // FrameImport FrameIncarnate
		reg,             // LayerRegister
		replRef,         // FrameImport
		smallInterface,  // FrameImport FrameIncarnate
		tooManySAPs,     // (nothing)
		xRtsInvoke       // CommInvoke
    };

#if RTS_COMPATIBLE <= 520
    typedef Error       rtsErrors; // for backwards compatibility
#endif

    Error               canPlugIn( const RTActorClass          *,
				   	   	   	   	   const RTComponentDescriptor * );

private:
    Error               errorNumber;

public:
    /** Get the most recent error that occurred in the controller's thread. Call this function
     immediately after a function has returned an error code, to get more information about the
     error that occurred.
     @return The error code for the most recent error that occurred in the controller's thread
     */
    RTS_INLINE Error    getError( void ) const;

    /** Get an error message for the most recent error that occurred in the controller's thread.
     Call this function immediately after a function has returned an error code, to get more
     information about the error that occurred.
	 @return The error message for the most recent error that occurred in the controller's thread
	 */
    const char        * strerror( void ) const;

    /** Print an error message to stderr. The printed messages will consist of a prefix followed
     by the message of the most recent error that occurred in the controller's thread.
     @param s A string that will precede the error message.
	 */
    void                perror( const char * s = "error" );

#if OTRTSDEBUG == DEBUG_NONE
    RTS_INLINE
#endif

	/** Set the most recent error for the controller's thread. This function is called by
	 TargetRTS functions when they fail. If you develop extensions to the TargetRTS your functions
	 may call it too.
	 @param error Error code
	 @param primitive Identifies the primitive action that failed
	 @param info A context object associated with the error (may be NULL)
	 @param detail An additional detailed context object associated with the error (may be NULL)
	 */
	void                setError( Error error,
								  Primitive primitive,
								  const void * info,
								  const void * detail);

#if OTRTSDEBUG < DEBUG_VERBOSE
    RTS_INLINE
#endif
    void                trace( Primitive, const void *, const void * );
#if OBSERVABLE
    void                breakPoint( RTActor *, RTEventInfo * );
    void                evtMsgQueue( void );

    /**
       Report all incoming messages satisfying actorPath in daemonInfo
     */
    void                getMessages( const RTDaemonInfo & daemonInfo );
#endif

	/** Get the name of the controller, which is the name of its physical thread (as defined in the
	 transformation configuration).
	 @return The name of the controller
	 */
    virtual const char * name( void ) const override;

    /** Kill this controller. All capsule instances that it runs will terminate. Messages that are
     pending to be dispatched will be deleted. The controller's thread will then terminate (if it
     is the main thread the whole application will terminate).
     */
    virtual void        kill( void ) override;

    /** Print various statistics to stdout about what has happened in the RT application (e.g.
     the number of incarnated capsules, the number of dispatched messages etc.).
	 */
    virtual void        printStats( void ) const;

    /** Abort the application. Calling this function is equivalent to calling kill() on the main
	 thread's controller.
	 */
    void                abort( void );

    /** Get the debugger associated with this controller, if any.
     @return The controller's debugger (NULL if none).
     */
    RTS_INLINE RTDebugger * debugger( void );

    void                destroy( RTActor * actor, RTActor * owner );

    int                 send( RTActor              * receiver,
			      	  	  	  int                    controlSignal,
							  const void           * data,
							  const RTObject_class * type,
							  int                    priority );

    RTS_INLINE void     receive( RTMessage *, RTController * sender );

    // Receive a copy of the message
    RTS_INLINE int      receiveMsgCopy(const RTMessage*, RTController * sender);

#if USE_THREADS
    void                peer_receive( RTMessage * );
#endif

    void                requeueMsg( RTMessage *, int front );

    RTS_INLINE RTMessage * newMsg( void );
    RTS_INLINE void     freeMsg( RTMessage * );

    RTTimerNode       * newTimerNode( void );
    void                freeTimerNode( RTTimerNode * node );

protected:
    // public to implementation

    // values returned by dispatch
    enum
    {
    	OneDelivered,
#if OTRTSDEBUG != DEBUG_NONE
		PermissionDenied,
#endif
		NoneToDeliver
    };

    int                 dispatch( int priorityLimit );
    void                awaitResponse( void );
    virtual void        waitForEvents( void ) = 0;
#if USE_THREADS
    virtual void        wakeup( void );
#endif

#if USE_THREADS
    RTSyncObject      * _sync;
#endif

private:
    static const unsigned char topPriority[ 1 << OTRTS_NUMPRIO ];

    RTMessageQ          internalQ[ OTRTS_NUMPRIO ];

#if USE_THREADS
    RTMessageQ          incomingQ[ OTRTS_NUMPRIO ];
#endif

    short               internalBits;
#if USE_THREADS
    short               incomingBits;
#endif

    RTMessage         * freeList;
    unsigned            freeListSize;

    RTTimerNode       * freeTimerList;
    unsigned            freeTimerListSize;

    // invoke reply messages are stored here
    RTMessage         * replyBuffer;

#if ! DEFER_IN_ACTOR
    RTMessageQ          deferredQ; // Deferred Messages Queue
#endif

    void                filterMessages( RTActor * receiver, RTQFilter & );
#if USE_THREADS
    void                acceptIncoming( void );
#endif

    static RTResourceMgr resourceMgr;

public:
    RTS_INLINE RTMessage * getReplyBuffer( void );
    RTS_INLINE RTMessage * setReplyBuffer( RTMessage * );
    RTS_INLINE void        restoreReplyBuffer( RTMessage * );

#if RTS_COUNT
    void                countActorDestroy   ( void );
    void                countActorIncarnate ( void );
    void                countMessageTransfer( RTMessage * );
    void                countPortDestroy    ( int );
    void                countPortIncarnate  ( int );
    void                countTimerCancel    ( void );
    void                countTimerExpire    ( void );
    void                countTimerStart     ( void );

protected:
    RTCounts            counts;
#else
    RTS_INLINE void     countActorDestroy   ( void );
    RTS_INLINE void     countActorIncarnate ( void );
    RTS_INLINE void     countMessageTransfer( RTMessage * );
    RTS_INLINE void     countPortDestroy    ( int );
    RTS_INLINE void     countPortIncarnate  ( int );
    RTS_INLINE void     countTimerCancel    ( void );
    RTS_INLINE void     countTimerExpire    ( void );
    RTS_INLINE void     countTimerStart     ( void );
#endif

private:
    // unavailable methods
    RTController( const RTController & );
    RTController & operator=( const RTController & );
}; //lint !e1712

#if RTS_INLINES
#include <RTController.inl>
#endif

#endif // __RTController_h__
