/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTExternal_h__
#define __RTExternal_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTInSignal_h__
#include <RTInSignal.h>
#endif

#ifndef __RTProtocol_h__
#include <RTProtocol.h>
#endif

struct RTExternalData;

/** Represents the UML-RT "External" protocol
 */
struct External
{
	/** Represents a port through which an external thread can safely communicate with its owner
	 capsule.
	 */
	class Base : public RTProtocol
	{
	public:
		static const RTProtocolDescriptor rt_class;

		RTS_INLINE Base(void);
		~Base(void);

		enum
		{
			rti_event = rtiLast_RTProtocol + 1
		};

		/** Enable the owner capsule to receive an event on the external port.
		 This function may only be called from the thread on which the owner capsule executes.
		 */
		void enable(void);

		/** Disable the owner capsule from receiving an event on the external port.
		 This function may only be called from the thread on which the owner capsule executes.
		 */
		void disable(void);

		/** Send an event on the external port. Data can optionally be included in the event.
		 This function may be used only on external threads (i.e. another thread than the one on
		 which the owner capsule executes). Sending of the event will only be successful if the
		 external port is currently enabled for event reception. After this function has been
		 called, the port is automatically disabled from receiving another event until enable() has
		 been called again.
		 @param[in] data Data object to put in the sent event. The data is copied by the TargetRTS
		 using the copy function of the type descriptor.
		 @param[in] type Type descriptor for the data object.
		 @return 1 on success and 0 if the event could not be sent on the external port (typically
		 because it was not enabled).
		 */
		int raise(const void * data = nullptr, const RTObject_class * type = nullptr);

		/** Store a data object on the external port. The data object will be put as the last object
		 in the list of objects stored on the external port. This function should normally be called
		 from an external thread when it has data that should be passed to a capsule thread, and
		 when it's not appropriate to pass the data in the call of raise(), for example to avoid
		 copying the data object.
		 @param[in] data Data object (non-null and dynamically allocated by code in the external
		 thread, which calls this function)
		 */
		void dataPushBack(void* data);

		/** Pop the first data object off the list of objects stored on the external port.
		 This function should normally be called by the thread on which the owner capsule executes.
		 It is the responsibility of the caller to delete the popped data object when done with it.
		 @param[out] data The address where to store the retrieved object. Will be null in case of
		 errors (for example when attempting to pop from an empty list).
		 @return The number of data objects remaining in the list of objects stored on the external
		 port.
		 */
		unsigned int dataPopFront(void** data);

		/** Pop the last data object off the list of objects stored on the external port.
		 This function should normally be called by the thread on which the owner capsule executes.
		 It is the responsibility of the caller to delete the popped data object when done with it.
		 @param[out] data The address where to store the retrieved object. Will be null in case of
		 errors (for example when attempting to pop from an empty list).
		 @return The number of data objects remaining in the list of objects stored on the external
		 port.
		 */
		unsigned int dataPopBack(void** data);

		/** Attempt to find a specific data object in the list of objects stored on the external
		 port. If it is found it will be deleted.
		 This function will typically be called from the same external thread that pushed the object
		 into the list, and it can be useful if the object contains data that has become invalid so
		 that it no longer should be processed by the owner capsule.
		 @param[in] data Data object to delete.
		 */
		void dataDelete(void* data);

		/** Return the object that represents the event which can be sent to the external port by an
		 external thread, and received by the thread on which the owner capsule executes.
		 */
		RTS_INLINE RTInSignal event(void);

	protected:
		enum
		{
			rtiLast_External = rti_event
		};

	private:
		RTMessage * msg;

		RTMessage * unload(void);

		static const RTSignalDescriptor rt_signals[];

		RTExternalData* externalData;

		// unavailable methods
		Base(const Base &);
		Base & operator=(const Base &);
	};
};

#if RTS_INLINES
#include <RTExternal.inl>
#endif

#endif // __RTExternal_h__
