/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTFrame_h__
#define __RTFrame_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTActorId_h__
#include <RTActorId.h>
#endif

#ifndef __RTController_h__
#include <RTController.h>
#endif

#ifndef __RTProtocol_h__
#include <RTProtocol.h>
#endif

struct RTActorClass;
class  RTActorRef;
class  RTController;
struct RTTypedValue;

/** Represents the UML-RT "Frame" protocol */
struct Frame
{
	/** Represents a frame service port */
    class Base : public RTProtocol
    {
    public:
		RTS_INLINE      Base( void );
		RTS_INLINE      ~Base( void );

		static const RTProtocolDescriptor rt_class;

		/** Incarnates a capsule by creating an instance of the capsule in an optional capsule part.
		 The capsule to incarnate is determined from the type of the capsule part.
		 @param[in] cp The capsule part where to insert the created capsule instance.
		 @return An RTActorId representing the created capsule instance.
		 */
		RTActorId       incarnate( RTActorRef & cp);

		/** Incarnates a capsule by creating an instance of the capsule in an optional capsule part.
		 @param[in] cp The capsule part where to insert the created capsule instance.
		 @param[in] cap The capsule to incarnate.
		 @return An RTActorId representing the created capsule instance.
		 */
		RTActorId       incarnate( RTActorRef & cp, const RTActorClass & cap);

		/** Incarnates a capsule by creating an instance of the capsule in an optional capsule part.
		 The capsule to incarnate is determined from the type of the capsule part.
		 @param[in] cp The capsule part where to insert the created capsule instance.
		 @param[in] data Initialization data to pass to the created capsule instance. It can be
		 accessed from the initial transition of its state machine.
		 @param[in] type Type descriptor for the data object.
		 @param[in] controller The controller which should run the created capsule instance. If it
		 is NULL, the same controller that runs the container capsule instance will be used.
		 @param[in] index The index within a replicated capsule part where to insert the created
		 capsule instance. Use -1 to insert the instance at the first free slot.
		 @return An RTActorId representing the created capsule instance.
		 */
		RTActorId       incarnate( RTActorRef & cp,
					   	   	   	   const void * data,
								   const RTObject_class * type,
								   RTController * controller,
								   int index );

		/** Incarnates a capsule by creating an instance of the capsule in an optional capsule part.
		 @param[in] cp The capsule part where to insert the created capsule instance.
		 @param[in] cap The capsule to incarnate.
		 @param[in] data Initialization data to pass to the created capsule instance. It can be
		 accessed from the initial transition of its state machine.
		 @param[in] type Type descriptor for the data object.
		 @param[in] controller The controller which should run the created capsule instance. If it
		 is NULL, the same controller that runs the container capsule instance will be used.
		 @param[in] index The index within a replicated capsule part where to insert the created
		 capsule instance. Use -1 to insert the instance at the first free slot.
		 @return An RTActorId representing the created capsule instance.
		 */
		RTActorId       incarnate( RTActorRef & cp,
					   	   	   	   const RTActorClass & cap,
								   const void * data,
								   const RTObject_class * type,
								   RTController * controller,
								   int index );

		/** Incarnates a capsule by creating an instance of the capsule in an optional capsule part.
		 The capsule to incarnate is determined from the type of the capsule part.
		 @param[in] cp The capsule part where to insert the created capsule instance.
		 @param[in] data Data object to pass to the created capsule instance. The data can be
		 accessed from the initial transition of its state machine.
		 @param[in] controller The controller which should run the created capsule instance. If it
		 is NULL, the same controller that runs the container capsule instance will be used.
		 @param[in] index The index within a replicated capsule part where to insert the created
		 capsule instance. Use -1 to insert the instance at the first free slot.
		 @return An RTActorId representing the created capsule instance.
		 */
		RTActorId       incarnate( RTActorRef & cp,
					   	   	   	   const RTDataObject & data,
								   RTController * controller = nullptr,
								   int index = -1 );

		/** Incarnates a capsule by creating an instance of the capsule in an optional capsule part.
		 @param[in] cp The capsule part where to insert the created capsule instance.
		 @param[in] cap The capsule to incarnate.
		 @param[in] data Data object to pass to the created capsule instance. The data can be
		 accessed from the initial transition of its state machine.
		 @param[in] controller The controller which should run the created capsule instance. If it
		 is NULL, the same controller that runs the container capsule instance will be used.
		 @param[in] index The index within a replicated capsule part where to insert the created
		 capsule instance. Use -1 to insert the instance at the first free slot.
		 @return An RTActorId representing the created capsule instance.
		 */
		RTActorId       incarnate( RTActorRef & cp,
					   	   	   	   const RTActorClass & cap,
								   const RTDataObject & data,
								   RTController * controller = nullptr,
								   int index = -1 );

		/** Incarnates a capsule by creating an instance of the capsule in an optional capsule part.
		 The capsule to incarnate is determined from the type of the capsule part.
		 @param[in] cp The capsule part where to insert the created capsule instance.
		 @param[in] info An RTTypedValue object which encapsules the data to pass to the created
		 capsule instance. The data can be accessed from the initial transition of its state machine.
		 @param[in] controller The controller which should run the created capsule instance. If it
		 is NULL, the same controller that runs the container capsule instance will be used.
		 @param[in] index The index within a replicated capsule part where to insert the created
		 capsule instance. Use -1 to insert the instance at the first free slot.
		 @return An RTActorId representing the created capsule instance.
		 */
		RTActorId       incarnate( RTActorRef & cp,
					   	   	   	   const RTTypedValue & info,
								   RTController * controller = nullptr,
								   int index = -1 );

		/** Incarnates a capsule by creating an instance of the capsule in an optional capsule part.
		 @param[in] cp The capsule part where to insert the created capsule instance.
		 @param[in] cap The capsule to incarnate.
		 @param[in] info An RTTypedValue object which encapsules the data to pass to the created
		 capsule instance. The data can be accessed from the initial transition of its state machine.
		 @param[in] controller The controller which should run the created capsule instance. If it
		 is NULL, the same controller that runs the container capsule instance will be used.
		 @param[in] index The index within a replicated capsule part where to insert the created
		 capsule instance. Use -1 to insert the instance at the first free slot.
		 @return An RTActorId representing the created capsule instance.
		 */
		RTActorId       incarnate( RTActorRef & cp,
					   	   	   	   const RTActorClass & cap,
								   const RTTypedValue & info,
								   RTController * controller = nullptr,
								   int index = -1 );

		/** Incarnates a capsule by creating an instance of the capsule in an optional capsule part using
		 a custom capsule factory for creating the capsule instance. Use this instead of the standard incarnate function
		 if you need to create the capsule instance in some special way, for example using
		 another capsule constructor than the one that is used by default.
		 @param[in] cp The capsule part where to insert the created capsule instance.
		 @param[in] factory The capsule factory which will be used for creating the capsule instance.
		 @param[in] index The index within a replicated capsule part where to insert the created
		 capsule instance. Use -1 to insert the instance at the first free slot.
		 @return An RTActorId representing the created capsule instance.
		 */
		RTActorId       incarnateCustom( RTActorRef & cp,
										 RTActorFactory& factory,
										 int index = -1);
		RTActorId       incarnateCustom( RTActorRef & cp,
										 RTActorFactory&& factory, // To allow temporary factory objects
										 int index = -1);

		/** Incarnates a capsule by creating an instance of the capsule in an optional capsule part using
		 a custom capsule factory for creating the capsule instance. Use this instead of the standard incarnate function
		 if you need to create the capsule instance in some special way, for example using
		 another capsule constructor than the one that is used by default.
		 @param[in] cp The capsule part where to insert the created capsule instance.
		 @param[in] data Initialization data to pass to the created capsule instance. It can be
		 accessed from the initial transition of its state machine.
		 @param[in] type Type descriptor for the data object.
		 @param[in] factory The capsule factory which will be used for creating the capsule instance.
		 @param[in] index The index within a replicated capsule part where to insert the created
		 capsule instance. Use -1 to insert the instance at the first free slot.
		 @return An RTActorId representing the created capsule instance.
		 */
		RTActorId       incarnateCustom( RTActorRef & cp,
				 	 	 	 	 	 	 const void * data,
										 const RTObject_class * type,
										 RTActorFactory& factory,
										 int index = -1);
		RTActorId       incarnateCustom( RTActorRef & cp,
				 	 	 	 	 	 	 const void * data,
										 const RTObject_class * type,
										 RTActorFactory&& factory, // To allow temporary factory objects
										 int index = -1);

		/** Destroy a capsule instance.
		 @param[in] i The capsule instance to destroy. It must be located in an optional capsule
		 part within the same capsule that owns the frame service port.
		 @return 1 if the capsule instance was successfully destroyed, otherwise 0.
		 */
		int             destroy( RTActorId  & i);

		/** Destroy all capsule instances contained in a certain capsule part.
		 @param[in] cp The capsule part. It must be optional and located within the same capsule
		 that owns the frame service port.
		 @return The number of destroyed capsule instances. 0 is returned in case of errors.
		 */
		int             destroy( RTActorRef & cp);

		/** Import a capsule instance into a plug-in capsule part.
		 @param[in] i Capsule instance to import. It must be located in an optional capsule part
		 within the same capsule that owns the frame service port.
		 @param[in] cp_dest The destination capsule part into which the capsule instance will be
		 imported. It must be a plug-in capsule part typed by a capsule that is compatible with the
		 type of the imported capsule instance.
		 @param[in] index The index within a replicated capsule part where to insert the imported
		 capsule instance. Use -1 to insert the instance at the first free slot.
		 @return 1 if the capsule instance was successfully imported, 0 otherwise.
		 */
		int             import( const RTActorId & i,
								RTActorRef & cp_dest,
								int index = -1 );

		/** Import a capsule instance into a plug-in capsule part.
		 @param[in] cp_src An optional capsule part that contains exactly one capsule instance that
		 should be imported into another plug-in capsule part. The source capsule part must be
		 located within the same capsule that owns the frame service port.
		 @param[in] cp_dest The destination capsule part into which the capsule instance will be
		 imported. It must be a plug-in capsule part typed by a capsule that is compatible with the
		 type of the imported capsule instance.
		 @param[in] index The index within a replicated capsule part where to insert the imported
		 capsule instance. Use -1 to insert the instance at the first free slot.
		 @return 1 if the capsule instance was successfully imported, 0 otherwise.
		 */
		int             import( RTActorRef & cp_src,
								RTActorRef & cp_dest,
								int index = -1 );

		/** Remove a capsule instance from a plug-in capsule part.
		 @param[in] i The capsule instance to remove (deport).
		 @param[in] cp The plug-in capsule part where the capsule instance is currently located.
		 @return 1 if the capsule instance was successfully deported, 0 otherwise.
		 */
		int             deport( const RTActorId & i,
								RTActorRef & cp);

		/** Return all capsule instances contained in a certain capsule part.
		 @param[in] cp A capsule part.
		 @return An array of pointers to the capsule instances that are contained in the capsule
		 part. The first and last element of the array is NULL, which means that you should use
		 1-based indexing for accessing a specific capsule instance. Note that it is the caller's
		 responsibility to delete the returned array. If you only want to access one capsule
		 instance it is therefore better to use incarnationAt().
		 */
		RTActorId    ** incarnationsOf( const RTActorRef & cp );

		/** Return a capsule instance contained in a certain capsule part.
		 @param[in] cp A capsule part.
		 @param[in] index The zero-based index of the capsule instance within the capsule part.
		 */
		RTActorId       incarnationAt ( const RTActorRef & cp,
										int index );

		/** Return the capsule instance that owns the frame service port.
		 @return The context capsule instance.
		 */
		RTActorId       me( void );

		/** Get the type of the capsule instance that owns the frame service port.
		 @return The capsule that is the type of the context capsule instance.
		 */
		const RTActorClass & myClass( void );

		/** Get the type of a certain capsule instance.
		 @param[in] instance A capsule instance
		 @return The capsule that is the type of the capsule instance, or EmptyActorClass in case
		 an error occurs.
		 */
		const RTActorClass & classOf( const RTActorId & instance );

		/** Determine if a capsule is the same or inherits from another capsule.
		 @param[in] subClass A capsule to compare the type of.
		 @param[in] superClass Another capsule to compare against.
		 @return 1 if subClass == superClass, or if subClass inherits from superClass. Otherwise 0
		 is returned.
		 */
		int classIsKindOf( const RTActorClass & subClass,
				       	   const RTActorClass & superClass );

		/** Get the name of a capsule type.
		 @return The name of the capsule as defined in the model.
		 */
		const char * className( const RTActorClass & );

    private:
#if RTFRAME_CHECKING != RTFRAME_CHECK_NONE
		int             permissionCheck( RTController::Primitive,
										 RTActorRef *,
										 int trace );
#endif

		// unavailable methods
		Base( const Base & );
		Base & operator=( const Base & );
    };
};

#if RTS_INLINES
#include <RTFrame.inl>
#endif

#endif // __RTFrame_h__
