/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTInSignal_h__
#define __RTInSignal_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTConfig_h__
#include <RTConfig.h>
#endif

class RTProtocol;

// It is common to make temporaries of this struct.
//lint -esym(1901,RTInSignal)

/** Provides the API for working with messages of a certain protocol event received on a certain
 port. Each protocol event is translated to a function on a class that inherits from RTProtocol, and
 such a function returns an RTInSignal object that is bound to a specific port and protocol event.
 This is true if the protocol event is incoming, or if it is outgoing and the port is conjugated.
 Call functions on that object to perform actions on one or several messages that have been received
 on the specified port, and which have the specified protocol event.
 */
struct RTInSignal
{
    // general application use

	/** Remove from the defer queue all messages that match the receiver port and the signal id of
	 the protocol event. If the receiver port is replicated messages will be deleted from the defer
	 queue regardless of the port index on which they were received.
	 @return The number of messages deleted from the defer queue.
	 */
    RTS_INLINE int purge      ( void );

	/** Remove from the defer queue all messages that match the receiver port and the signal id of
	 the protocol event. Only the messages that were received on a certain port index will be
	 deleted from the defer queue.
	 @param[in] index Port index (0-based)
	 @return The number of messages deleted from the defer queue.
	 */
    RTS_INLINE int purgeAt    ( int index );

    /** Recall a message from the defer queue and insert it in the controller's message
	 queue (either before or after other queued messages). At most one message will be recalled and
	 it must match the receiver port and the signal id of the protocol event (however, it doesn't
	 matter on which port index it was received).
	 @param[in] front If 0 the recalled message will be inserted at the back of the message queue,
	 otherwise it will be inserted at the front of the message queue.
	 @return The number of messages recalled from the defer queue (i.e. either 1 or 0).
	 */
    RTS_INLINE int recall     ( int front = 0 );

    /** Recall a message from the defer queue and insert it in the controller's message
	 queue (either before or after other queued messages). At most one message will be recalled and
	 it must match the receiver port and the signal id of the protocol event. It must also have
	 been received on a certain port index.
	 @param[in] index Port index (0-based)
	 @param[in] front If 0 the recalled message will be inserted at the back of the message queue,
	 otherwise it will be inserted at the front of the message queue.
	 @return The number of messages recalled from the defer queue (i.e. either 1 or 0).
	 */
    RTS_INLINE int recallAt   ( int index, int front = 0 );

    /** Recall one or many messages from the defer queue and insert them in the controller's message
	 queue (either before or after other queued messages). All messages will be recalled that
	 match the receiver port and the signal id of the protocol event (however, it doesn't
	 matter on which port index they were received).
	 @param[in] front If 0 the recalled messages will be inserted at the back of the message queue,
	 otherwise they will be inserted at the front of the message queue.
	 @return The number of messages recalled from the defer queue.
	 */
    RTS_INLINE int recallAll  ( int front = 0 );

    /** Recall one or many messages from the defer queue and insert them in the controller's message
	 queue (either before or after other queued messages). All messages will be recalled that
	 match the receiver port and the signal id of the protocol event. They must also have
	 been received on a certain port index.
	 @param[in] index Port index (0-based)
	 @param[in] front If 0 the recalled messages will be inserted at the back of the message queue,
	 otherwise they will be inserted at the front of the message queue.
	 @return The number of messages recalled from the defer queue.
	 */
    RTS_INLINE int recallAllAt( int index, int front = 0 );

    // for use by run-time services library only
    RTS_INLINE RTInSignal( RTProtocol *, int );
    RTS_INLINE ~RTInSignal( void );

    /** The port on which the message is received. */
    RTProtocol * _port;

    /** The signal id of the protocol event for the received message. */
    int          _signal;
}; //lint !e1712

#if RTS_INLINES
#include <RTInSignal.inl>
#endif

#endif // __RTInSignal_h__
