/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTInjector_h__
#define __RTInjector_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTConfig_h__
#include <RTConfig.h>
#endif

#ifndef RTUseSTL
#error RTInjector requires STL support, but RTUseSTL is not defined.
#endif // RTUseSTL

#include <string>
#include <functional>
#include <unordered_map>

class RTController;
class RTActorRef;
class RTActor;
class RTMutex;

/**
 * Dependency injector for RT applications. At runtime a capsule instance depends on other
 * capsule instances. Instead of hard coding such dependencies inside the capsule itself,
 * the dependency injector approach allows to register such dependencies in a centralized location.
 * Thereby you can configure the application by changing in that single place only.
 *
 * A typical usage of this class is to access the singleton instance at application start-up (for example in
 * the constructor of the top capsule). There you then configure the dependencies to be injected by registering
 * create functions for capsule parts that you want to incarnate in a non-default way. Some example of
 * customization possibilities (implemented by the create function):
 * - which capsule to create an instance of (if not the same as specified as type for the capsule part)
 * - which constructor arguments to pass (custom initialization data for the create capsule instance)
 * - which thread to run the created capsule instance (if not the same as the one running the owner capsule instance)
 * A capsule factory (local or global in the TC) can be used and delegate to the injector object whenever
 * a capsule instance should be created. The injector will check if a specific create function has been registered
 * as a dependency, and if so it will use it for creating the capsule instance. If no such create function is found
 * it will fallback to creating the capsule instance as usual (based on the type of the capsule part).
 *
 * Dependencies are usually not changed once registered, but if needed it's possible to do so (the implementation is thread safe).
 */
class RTInjector
{
public:

	/**
	 * Access the singleton injector instance (created on first access)
	 * @return injector instance
	 */
	static RTInjector& getInstance() {
		static RTInjector instance;
		return instance;
	}

	RTInjector();
	~RTInjector();

	/**
	 * Register a create function to be used for incarnating capsule instances in a capsule part.
	 * @param id Qualified id string for the capsule part (e.g. "/part1:0/part2")
	 * @param createFunc Create function to register
	 */
	void registerCreateFunction(const std::string& id, std::function<RTActor* (RTController *, RTActorRef *, int)> createFunc);

	/**
	 * Get a create function registered for incarnating capsule instances in a capsule part.
	 * @param id Qualified id string for the capsule part (e.g. "/part1:0/part2")
	 * @return Registered create function, or an empty function if no create function was registered for the capsule part
	 */
	std::function<RTActor* (RTController *, RTActorRef *, int)> getCreateFunction(const std::string& id);

	/**
	 * Create a capsule instance. If a create function has been registered for the specified capsule part it will be used.
	 * Otherwise, the default create function of the capsule will be used.
	 * @param rts Default controller thread to run the created capsule instance (a registered create function may override this)
	 * @param ref Capsule part to host the created capsule instance
	 * @param index Index where to insert the created capsule instance in the capsule part
	 * @return Capsule instance, or nullptr in case a capsule instance couldn't be created
	 */
	RTActor* create(RTController *rts, RTActorRef *ref, int index);

	RTInjector(RTInjector const&) = delete;
	void operator=(RTInjector const&) = delete;

private:

	std::unordered_map<std::string, std::function<RTActor* (RTController *, RTActorRef *, int)> > createFuncRegistry;
#if USE_THREADS
	RTMutex* _mutex;
#endif

};

#endif // __RTInjector_h__
