/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTMessage_h__
#define __RTMessage_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTTimespec_h__
#include <RTTimespec.h>
#endif

class  RTActor;
class  RTController;
class  RTDataObject;
struct RTObject_class;
class  RTProtocol;
struct RTTypedValue;

/** Represents a message used for communication between capsule instances. The message may carry
 a data object. A message object is usually an instance of a protocol event defined in the model.
 However, there are also internal messages used by the TargetRTS itself.
 */
class RTMessage
{
public: // these really should be private
    RTMessage             * next;

#if OTRTSDEBUG != DEBUG_NONE
    int                     _sendPortId;
    int                     _sendPortIndex;
    int                     _sendProbeId;
    RTTimespec              _sendTime;
#endif

public:
    union
    {
		void              * data;
		const void        * const_data;
    };
    short                   signal;
    unsigned char           priority;

protected:
    unsigned char           _flags;
    RTActor               * _receiver;
    RTProtocol            * _port;
    int                     _portIndex;
    const RTObject_class  * _type;

#if RTMESSAGE_PAYLOAD_SIZE != 0
    // space for a small amount of data
    union
    {
		char                _data_area[ RTMESSAGE_PAYLOAD_SIZE ];
		RTAlignment         _alignment;
    };
#endif

    void                    _internData( bool moveData );

    enum
    {
		MaskDeferred  = 1,
		MaskDataOwned = 2
    };

public:
    /** Destroy the data object stored in the message, and reset the message object to make it
     ready to be used in another communication. This function is called internally by the TargetRTS
     and there is usually no need to call it directly.
     */
    RTS_INLINE void         clear( void );

    RTS_INLINE              RTMessage( void );
			    			RTMessage( const RTMessage & );

    RTS_INLINE              ~RTMessage( void );

    /** Determines if the message carries a data object which is an RTDataObject.
     Mainly provided for backwards compatibility.
     @return 1 if yes, 0 if no.
     */
    RTS_INLINE int          isRTData( void ) const;

    /** Mark this message as deferred. This function is called internally by the TargetRTS
     and there is usually no need to call it directly.
     */
    RTS_INLINE void         setDeferred( void );

    /** Mark this message as not deferred. This function is called internally by the TargetRTS
	 and there is usually no need to call it directly.
	 */
    RTS_INLINE void         resetDeferred( void );

    /** Determines if this message has been deferred, i.e. placed in a defer queue.
     @return 1 if yes, 0 if no.
     */
    RTS_INLINE int          isDeferred( void ) const;

    RTS_INLINE void         setInvalid( void );

    /** A message is considered valid as soon as its protocol event (i.e. signal id) has been setup.
     @return 1 if the message has a protocol event setup, 0 otherwise.
     */
    RTS_INLINE int          isValid( void ) const;

    /** Set the signal id of the message, i.e. the id of its the protocol event.
     @param[in] id Numeric id of the protocol event of the message.
     */
    RTS_INLINE void         setSignal( int id);

    /** Get the signal id of the message, i.e. the id of its protocol event.
     @return Numeric id of the protocol event of the message.
     */
    RTS_INLINE int          getSignal( void ) const;

    /** Get the signal name of the message, i.e. the name of its protocol event.
	 @return Name of the protocol event of the message.
	 */
    const char            * getSignalName( void ) const;

    /** Set the priority of the message. See RTPriority.h for a list of valid message priorities.
	 @param[in] prio Priority for the message.
	 */
    RTS_INLINE void         setPriority( int prio);

    /** Get the priority of the message. See RTPriority.h for a list of valid message priorities.
	 @return The priority for the message.
	 */
    RTS_INLINE int          getPriority( void ) const;

    /** Get the data object carried by the message, if any.
     @return The data object carried by the message, or NULL if none.
     */
    RTS_INLINE void       * getData( void ) const;

    /** Set the data object of the message.
	 @param[in] data Data object.
	 @param[in] type Type descriptor of the data object.
	 */
    RTS_INLINE void         setData( const void * data, const RTObject_class * type);

    /** Set the data object of the message.
	 @param[in] data Data object.
	 */
    RTS_INLINE void         setData( const RTDataObject & data);

    /** Intern the data object of the message. Normally this means the data object will be copied into the message
     but if moveData is true, the data object will instead be moved into the message.
     */
    RTS_INLINE void         internData( bool moveData = false );

    /** Get the type of the data object of the message, if any.
     @return Type descriptor for the data object, or NULL in case the message does not have one.
     */
    RTS_INLINE const RTObject_class * getType( void ) const;

    /** Return the capsule port on which this message was received.
     @return The port on which the message was received, or NULL if it was not yet received.
     */
    RTS_INLINE RTProtocol * sap( void ) const;

    /** Return the id of the capsule port on which this message was received.
	 @return The id of the port on which the message was received, or NULL if it was not yet
	 received.
	 */
    int                     sapId( void ) const;

    /** Return the index (0-based) of the replicated port on which this message was received.
     It's for example useful to get this index if you want to send a message back to the sender.
	 @return The 0-based index of the port on which the message was received.
	 */
    RTS_INLINE int          sapIndex0( void ) const;

    /** Return the index (1-based) of the replicated port on which this message was received.
	 It's for example useful to get this index if you want to send a message back to the sender.
	 @return The 1-based index of the port on which the message was received.
	 */
    RTS_INLINE int          sapIndex( void ) const;

    enum DeliverResult
    {
    	Accepted
      , Deferred
      , Rejected // signal or type invalid
    };

    /** Attempt to deliver the message to its specified receiver. This function is called by
     the TargetRTS when a message is sent, and it's normally not useful to call it directly.
     @return A delivery status informing about how the receiver handled the message.
     */
    DeliverResult           deliver( void );

    RTMessage             & operator=( const RTMessage & );

    /** Defer the message by placing it in the defer queue. You can later recall a deferred message
     by calling one of the recall functions of RTInSignal or RTProtocol.
     @return 1 if the message was successfully deferred, 0 otherwise.
     */
    int                     defer( void ) const;

    /** Set the destination (i.e. receiver) of the message. This function is called by
     the TargetRTS when a message is sent, and it's normally not useful to call it directly.
     @param[in] cap Receiver capsule instance.
     @param[in] port The receiver port on the capsule instance to which the message is sent.
     @param[in] index The 0-based index of the receiver port (in case it is replicated).
     */
    RTS_INLINE void         setDestination( RTActor * cap, RTProtocol * port, int index);

    /** Return the currently set receiver for the message.
     @return The capsule instance set-up as the receiver of the message.
     */
    RTS_INLINE RTActor    * receiver( void ) const;

    /** Return the controller of the message receiver.
     @return The controller of the capsule instance that is set-up as the receiver of the message.
     */
    RTS_INLINE RTController * context( void ) const;
};

#if RTS_INLINES
#include <RTMessage.inl>
#endif

#endif // __RTMessage_h__
