/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTObject_class_h__
#define __RTObject_class_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTFieldOffset_h__
#include <RTFieldOffset.h>
#endif

#ifndef __RTVersionId_h__
#include <RTVersionId.h>
#endif

struct RTFieldDescriptor;
struct RTObject_class;

typedef void * ( * RTSuperAccessFunction )( void * derived );

typedef void ( * RTInitFunction    )( const RTObject_class * type,
				      void                 * target );

typedef void ( * RTCopyFunction    )( const RTObject_class *,
				      void                 * target,
				      const void           * source );

typedef void ( * RTMoveFunction    )( const RTObject_class *,
				      void                 	* target,
				      void           		* source );

typedef void ( * RTDestroyFunction )( const RTObject_class * type,
				      void                 * target );

#if OBJECT_DECODE
    class RTDecoding;
    typedef int ( * RTDecodeFunction )( const RTObject_class * type,
					void                 * target,
					RTDecoding           * );
#endif

#if OBJECT_ENCODE
    class RTEncoding;
    typedef int ( * RTEncodeFunction )( const RTObject_class * type,
					const void           * source,
					RTEncoding           * );
#endif

/** A type descriptor providing information about a type. For most types defined in the model
 a static variable of this struct will be automatically generated. For more complex types, or to
 have a type descriptor with a non-default behavior, you can write you own custom type descriptor.
 Capsule types have their own type descriptor struct; see RTActor_class.
 */
struct RTObject_class
{
    enum DestroyOption
    {
    	DestroyOnly,
		DestroyAndDeallocate
    };

    // access functions

    /** Get the descriptor for the super type, i.e. the type from which the described type inherits.
     @return Type descriptor for the super type, or null if there is no super type.
     */
    RTS_INLINE const RTObject_class * super  ( void ) const;

    /** Get the name of the described type.
     @return Type name
     */
    RTS_INLINE const char * name ( void ) const;

    /** Get the version of the described type.
     @return Type version (usually 0 for the first version, and then incremented by one each time
     the type changes).
     */
    RTS_INLINE RTVersionId version( void ) const;

    /** Determines if the described type is the same, or a subtype of, another type.
     @param other Type descriptor for the other type to check the described type against.
     @return 1 if yes, 0 if no.
     */
    int isKindOf( const RTObject_class * other) const;

    /** Allocate new memory for holding an instance of the described type.
     @return Allocated memory big enough to hold an instance of the described type.
     */
    void * allocate( void ) const;

    /** Initialize an instance of the described type.
     @param[in] data Data object for an instance of the described type, typically allocated by a
     call to allocate().
     */
    RTS_INLINE void init( void * data) const;

    /** Make a copy of an instance of the described type.
     @param[in] to The target object into which the data will be copied.
     @param[in] from The source object that will be copied.
     */
    RTS_INLINE void copy( void * to, const void * from) const;

    /** Move an instance of the described type.
     @param[in] to The target object into which the data will be moved.
     @param[in] from The source object that will be moved from.
    */
    RTS_INLINE void move( void * to, void * from) const;

#if OBJECT_DECODE

    /** Decode an external representation (e.g. a string representation) and populate an instance of
     the described type accordingly.
     @param[in] data The target object into which the decoded data will be written.
     @param[in] decoder The decoder implementation.
     @return 1 if decoding was successful, 0 otherwise.
     */
    RTS_INLINE int decode( void * data, RTDecoding * decoder) const;
#endif
#if OBJECT_ENCODE

    /** Encode an instance of the described type into an external representation (e.g. a string
     representation).
     @param[in] data The instance of the described type to encode.
     @param[in] encoder The encoder implementation.
     @return 1 if encoding was successful, 0 otherwise.
     */
    RTS_INLINE int encode( const void * data, RTEncoding * encoder) const;
#endif

    /** Destroy an instance of the described type.
     @param[in] inst An instance of the described type to destroy.
     @param[in] opt Specifies whether the memory for the data object should be deallocated as part
     of destruction or not.
     */
    void destroy( void * inst, DestroyOption opt) const;

    /** Register the type descriptor so it can be looked up quickly using the name of the described
     type as the key. See lookup() for how to perform the lookup.
     All generated type descriptors will be automatically installed, so normally this function should
     not be called from user code. See RTTypeInstaller for more information.
     However, if you have manually written types for which you want to register a type descriptor, you need
     to call this function to install it in the registry.
     The name of the installed type is used as the key in the registry. If a type with that name is already registered
     (the same type or a different one) this function does nothing, but prints a warning.
     @param[in] warnForDuplicates Set to false to avoid printing a warning if trying to register a type descriptor with
     a name that is already used.
     */
    void                              install( bool warnForDuplicates = true ) const;
#if OBJECT_DECODE

    /** Lookup a type descriptor by the name of the type it describes.
     @param[in] className Name of the type.
     @return The type descriptor for the type, or null if the TYPE doesn't have a type descriptor defined.
     */
    static const RTObject_class     * lookup( const char * className );
#endif

    /** Get a type descriptor from a type TYPE at compile-time. This template function is specialized
     for all built-in and RT types. If you have manually implemented a type descriptor for a type you
     must also manually specialize this template if you need to lookup that type descriptor at compile-time.
     @return The type descriptor for the type, or null if the TYPE doesn't have a type descriptor defined.
     */
    template <typename TYPE> static const RTObject_class * fromType();

    // implementation

    /** The super type from which the type inherits */
    const RTObject_class            * _super;

    /** The super object accessor function for this type */
    RTSuperAccessFunction             _super_accessor;

    /** The name of the type */
    const char                      * _name;

    /** The version id of the type */
    RTVersionId                       _version;

    /** The size (in bytes) of the type */
    RTFieldOffset                     _size;

    // These function pointers must all be non-nil.

    /** The init function of the type, i.e. its default constructor */
    RTInitFunction                    _init_func;

    /** The copy function of the type, i.e. its copy constructor */
    RTCopyFunction                    _copy_func;

    /** The move function of the type, i.e. its move constructor */
    RTMoveFunction                    _move_func;


#if OBJECT_DECODE

    /** The decode function of the type */
    RTDecodeFunction                  _decode_func;
#endif
#if OBJECT_ENCODE

    /** The encode function of the type */
    RTEncodeFunction                  _encode_func;
#endif

    /** The destroy function of the type, i.e. its destructor */
    RTDestroyFunction                 _destroy_func;

    /** This variable is used for structured and array-like types to specify the number of
     fields or elements that are contained in the type. For other kinds of types this variable is
     not relevant. */
    int                               _num_fields;

    /** This variable is used for structured and array-like types to specify the type of fields
     or elements that are contained in the type. For other kinds of types this variable is
     not relevant. The variable is an array of descriptor objects, each of which describes one field
     or element of the type. */
    const RTFieldDescriptor         * _fields;
};

// Trivial (no-op) init, copy and destroy functions.
    extern void RTnop_init( const RTObject_class *, void * );
    extern void RTnop_copy( const RTObject_class *, void *, const void * );
    extern void RTnop_move( const RTObject_class *, void *, void * );
    extern void RTnop_destroy( const RTObject_class *, void * );

// Functions for instances describing struct or class types
    extern void RTstruct_init( const RTObject_class *, void * );
    extern void RTstruct_copy( const RTObject_class *, void *, const void * );
#if OBJECT_DECODE
    extern int  RTstruct_decode( const RTObject_class * type,
				 void                 * target,
				 RTDecoding           * );
#endif
#if OBJECT_ENCODE
    extern int  RTstruct_encode( const RTObject_class * type,
				 const void           * target,
				 RTEncoding           * );
#endif
    extern void RTstruct_destroy( const RTObject_class *, void * );

// Functions for instances describing enum types
#if OBJECT_DECODE
    extern int  RTenum_decode( const RTObject_class * type,
			       void                 * target,
			       RTDecoding           * );
#endif
#if OBJECT_ENCODE
    extern int  RTenum_encode( const RTObject_class * type,
			       const void           * target,
			       RTEncoding           * );
#endif

// Functions for instances which can inherit implementations
#if OBJECT_DECODE
    extern int  RTinherit_decode( const RTObject_class * type,
				  void                 * target,
				  RTDecoding           * );
#endif
#if OBJECT_ENCODE
    extern int  RTinherit_encode( const RTObject_class * type,
				  const void           * target,
				  RTEncoding           * );
#endif

// Functions for instances of RTObject_class describing arrays.
    extern void RTarray_init( const RTObject_class *, void * );
    extern void RTarray_copy( const RTObject_class *, void *, const void * );
#if OBJECT_DECODE
    extern int  RTarray_decode( const RTObject_class * type,
				void                 * target,
				RTDecoding           * );
#endif
#if OBJECT_ENCODE
    extern int  RTarray_encode( const RTObject_class * type,
				const void           * target,
				RTEncoding           * );
#endif
    extern void RTarray_destroy( const RTObject_class *, void * );

// Init and copy functions for abstract classes: these simply panic.
    extern void RTabstract_init( const RTObject_class *, void * );
    extern void RTabstract_copy( const RTObject_class *, void *, const void * );
    extern void RTabstract_move( const RTObject_class *, void *, void * );

// Reusable basic type descriptors. The convention is that for a type called T,
// one may refer to its type descriptor using the name RTType_T.
// In some cases the name RTType_T will refer to a macro.

    extern const RTObject_class RTType_void;
    extern const RTObject_class RTType_bool;
    extern const RTObject_class RTType_char;
#if RTUseFloatingPoint
    extern const RTObject_class RTType_double;
    extern const RTObject_class RTType_float;
#endif
    extern const RTObject_class RTType_int;
    extern const RTObject_class RTType_long;
    extern const RTObject_class RTType_short;
    extern const RTObject_class RTType_unsigned;

	extern const RTObject_class RTType_unsignedint;
	extern const RTObject_class RTType_unsignedlong;
	extern const RTObject_class RTType_unsignedshort;
	extern const RTObject_class RTType_unsignedchar;
    typedef char           * RTpchar;
    typedef void           * RTpvoid;
    typedef unsigned char    RTuchar;
    typedef unsigned long    RTulong;
    typedef unsigned short   RTushort;

    extern const RTObject_class RTType_RTpchar;
    extern const RTObject_class RTType_RTpvoid;
    extern const RTObject_class RTType_RTuchar;
    extern const RTObject_class RTType_RTulong;
    extern const RTObject_class RTType_RTushort;

    // Template specializations for mapping built-in types to type descriptors
    template <> inline const RTObject_class* RTObject_class::fromType<int>() { return &RTType_int; }
    template <> inline const RTObject_class* RTObject_class::fromType<char>() { return &RTType_char; }
    template <> inline const RTObject_class* RTObject_class::fromType<float>() { return &RTType_float; }
    template <> inline const RTObject_class* RTObject_class::fromType<bool>() { return &RTType_bool; }
    template <> inline const RTObject_class* RTObject_class::fromType<double>() { return &RTType_double; }
    template <> inline const RTObject_class* RTObject_class::fromType<short>() { return &RTType_short; }
    template <> inline const RTObject_class* RTObject_class::fromType<long>() { return &RTType_long; }
    template <> inline const RTObject_class* RTObject_class::fromType<unsigned int>() { return &RTType_unsignedint; } // Same as RTType_unsigned
    template <> inline const RTObject_class* RTObject_class::fromType<unsigned char>() { return &RTType_unsignedchar; } // Same as RTType_RTuchar
    template <> inline const RTObject_class* RTObject_class::fromType<unsigned long>() { return &RTType_unsignedlong; } // Same as RTType_RTulong
    template <> inline const RTObject_class* RTObject_class::fromType<unsigned short>() { return &RTType_unsignedshort; } // Same as RTType_RTushort
    template <> inline const RTObject_class* RTObject_class::fromType<char*>() { return &RTType_RTpchar; }


#if OBJECT_DECODE
class RTTypeInstaller
{
public:
    explicit    RTTypeInstaller( const RTObject_class & );

    static void installAll( void );

private:
    static RTTypeInstaller * _head;
    RTTypeInstaller        * _next;
    const RTObject_class   * _type;

    // unavailable methods
    RTTypeInstaller( const RTTypeInstaller & );
    RTTypeInstaller & operator=( const RTTypeInstaller & );
}; //lint !e1712
#endif // OBJECT_DECODE

#if RTS_INLINES
#include <RTObject_class.inl>
#endif

#endif // __RTObject_class_h__
