/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTProtocol_h__
#define __RTProtocol_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTPriority_h_
#include <RTPriority.h>
#endif

#ifndef __RTProtocolDescriptor_h__
#include <RTProtocolDescriptor.h>
#endif

class  RTActor;
class  RTActorProbe;
struct RTBindingEnd;
class  RTController;
class  RTDebugger;
struct RTExceptionSignal;
struct RTInSignal;
class  RTLayerConnector;
class  RTLayerData;
class  RTMessage;
struct RTObject_class;
struct RTOutSignal;
struct RTPortDescriptor;
struct RTSymmetricSignal;
struct RTTypedValue;

/** Represents a general capsule port typed by a protocol which determines the set of events that
can be sent to and received by the port.
All specific protocols inherit from RTProtocol, which means that the functions of this class
can be called on any capsule port regardless of which protocol it is typed by.
For all protocols defined in the model two subclasses of RTProtocol are generated; one for each
direction of communication using the events defined in the protocol. For a protocol X these
subclasses are X::Base and X::Conjugate.
 */
class RTProtocol
{
public:
    RTS_INLINE RTProtocol( void );
			   ~RTProtocol( void );

	/** Get the context of the capsule that owns this port.
	 @return The controller that runs the capsule instance to which this port belongs.
	 */
    RTController * context( void ) const;

    /** Get the id of the port.
     @return An id that is unique among all the ports of the owner capsule.
     */
    RTS_INLINE int getId( void ) const;

    /** Get information about a port.
     @return An RTPortDescriptor object that contains information about the port. This information
     is static and is constituted by port properties defined in the design model, such as the name
     of the port.
     */
    const RTPortDescriptor * getInfo( void ) const;

    /** Get the name of the port.
     @return The name of the port.
     */
    RTS_INLINE const char * getName( void ) const;

    /** Get the owner capsule to which the port belongs.
     @return The RTActor representing the capsule to which this port belongs.
     */
    RTS_INLINE RTActor * getOwner( void ) const;

	/** Get the size of the port.
	 @return The replication factor of the port, i.e. the size as defined by the port multiplicity
	 in the model. This specifies the maximum number of port instances, which is also the maximum
	 number of possible connections to remote port instances.
	 */
    RTS_INLINE int size( void ) const;

    /** Set a new size (i.e. multiplicity) for the port. This can only be done on non-wired ports.
     Increasing the size gives room for more connections, while decreasing the size may lead to that
     some established connections will be dropped.
     @param new_size New size for the port. Must be >= 0.
     @return 1 if successful and 0 if resizing fails (use RTActor::getError() then to determine the
     cause for the failure.
     */
    int resize( int new_size );

    /** Register an unwired port (SAP) with the layer service (as a "client"). If this function
     is invoked on an SAP which is already registered with a different name, then the original
     registered name is automatically deregistered, and the SAP is registered with the new name.
     Registering an SAP enables an SPP to later connect to it using the specified service name.
     @param service A string that is used to identify a unique service name under which SAPs and
     SPPs will connect.
	 @return 1 if the registration of the service name was successful, and 0 otherwise. The
	 registration fails if the port is not an unwired port. If the SAP is already registered
	 with this same name, the function returns 1.
     */
    RTS_INLINE int registerSAP( const char * service);

    /** Deregister an unwired port (SAP). If the port is currently connected to an SPP, the
     connection will be terminated.
     @return 1 if deregistration was successful, 0 otherwise.
     */
    RTS_INLINE int deregisterSAP( void );

    /** Register an unwired port (SPP) with the layer service (as the "provider"). If this
	 function is invoked on an SAP which is already registered with a different name, then the
	 original registered name is automatically deregistered, and the SAP is registered with the new
	 name. Registering an SPP enables SAPs to later connect to it using the specified service name.
	 @param service A string that is used to identify a unique service name under which SAPs and
	 SPPs will connect.
	 @return 1 if the registration of the service name was successful, and 0 otherwise. The
	 registration fails if the port is not an unwired port. If the SPP is already registered
	 with this same name, the function returns 1.
	 */
    RTS_INLINE int registerSPP( const char * service);

    /** Deregister an unwired port (SPP). If the port is currently connected to SPPs, all those
	 connections will be terminated.
	 @return 1 if deregistration was successful, 0 otherwise.
	 */
    RTS_INLINE int deregisterSPP( void );

    /** Get the default registration name for the port.
     @return The default registration name specified for the unwired port. This corresponds to
     the property "Registration Override" specified in the model.
     */
    const char * defaultLayerName( void ) const;

    /** Determine if the port is currently registered with the layer service or not.
     @return 1 if the unwired port is currently registered with the layer service, 0 otherwise.
     */
    RTS_INLINE int isRegistered( void ) const;

    /** Get the registration name of the port.
     @return The name under which the unwired port is currently registered with the layer
     service.
     */
    const char * getRegisteredName( void ) const;

#if RTS_COMPATIBLE <= 520
    /** Register an unwired port either as an SAP or SPP (depending on the "Publish" property
     set on the port in the model).
	 @deprecated Be specific and use registerSPP() or registerSAP() instead.
	 @return 1 on successful registration, 0 otherwise.
     */
    int registerAs( const char * service);
#endif

    /** Deregister an unwired port (either an SAP or SPP). This is equivalent of calling
     deregisterSAP() or deregisterSPP().
     @return 1 if deregistration was successful, 0 otherwise.
     */
    int deregister( void );

    /** Determine if a port instance is currently connected to another port instance.
     @param index 0-based index of the port instance within the port (0 <= index < size())
     @return 1 if the port instance at the specified index is connected to another port instance.
	 0 if it is not connected.
     */
    int isBoundAt( int index ) const;

    /** Determine if a port instance is currently connected to a port instance owned by a certain
	 capsule instance.
	 @param index 0-based index of the port instance within the port (0 <= index < size())
	 @param capsule Capsule instance
	 @return 1 if the port instance at the specified index is connected to another port instance
	 that is owned by the specified capsule instance. 0 if not.
	 */
    int isIndexTo( int index, RTActor * capsule) const;

    /** Determine if there exists a port instance currently connected to a port instance owned by a
     certain capsule instance. If so, the index of the first found matching port instance is
     returned.
     @param capsule Capsule instance
     @return The smallest index found for a port instance connected to another port instance that is
     owned by the specified capsule instance. -1 if no such port instance exists.
     */
    int indexTo( RTActor * capsule) const;

    // Managing the defer queue: These functions match any signal.

    /** Empty the defer queue of all port instances without recalling any deferred message. To
     delete deferred messages for one port instance only, use purgeAt().
     @return The number of messages deleted from the defer queue.
     */
    int purge( void );

    /** Empty the defer queue of a specified port instance without recalling any deferred message.
	 To delete deferred messages for all port instances, use purge().
	 @return The number of messages deleted from the defer queue.
	 */
    int purgeAt( int index );

    /** Recall a message from the defer queue and insert it at the back of the controller's message
     queue (after other queued messages). The recalled message can be for any port instance.
     @return The number of messages recalled from the defer queue (i.e. either 1 or 0).
     */
    int recall( void );

    /** Recall a message from the defer queue and insert it at the front of the controller's
	 message queue (before other queued messages). The recalled message can be for any port instance.
	 @return The number of messages recalled from the defer queue (i.e. either 1 or 0).
	 */
    int recallFront( void );

    /** Recall a message from the defer queue and insert it in the controller's message queue. By
	 default the recalled message is inserted at the back of the message queue. The recalled message
	 must be for the port instance at the specified index.
	 @param index Index of the port instance for which to recall a message.
	 @param front Set this parameter to 1 to insert the recalled message at the front of the message
	 queue instead of at the back which is the default behavior.
	 @return The number of messages recalled from the defer queue (i.e. either 1 or 0).
	 */
    int recallAt( int index, int front = 0 );

    /** Recall all messages from the defer queue and insert them at the back of the controller's
	 message queue (after other queued messages). The recalled messages can be for any port
	 instance.
	 @return The number of messages recalled from the defer queue.
	 */
    int recallAll( void );

    /** Recall all messages from the defer queue and insert them at the front of the controller's
	 message queue (before other queued messages). The recalled messages can be for any port
	 instance.
	 @return The number of messages recalled from the defer queue.
	 */
    int recallAllFront( void );

    /** Recall all messages from the defer queue and insert them in the controller's message queue.
	 By default the recalled messages are inserted at the back of the message queue. The recalled
	 messages must be for the port instance at the specified index.
	 @param index Index of the port instance for which to recall the messages.
	 @param front Set this parameter to 1 to insert the recalled messages at the front of the
	 message queue instead of at the back which is the default behavior.
	 @return The number of messages recalled from the defer queue.
	 */
    int recallAllAt( int index, int front = 0 );

    /** Turn on or off binding notifications for this port. When such notifications are turned on
     the port will receive the 'rtBound' and 'rtUnbound' messages each time a port instance gets
     connected to (bound) or disconnected from (unbound) a remote port instance. Note that no
     messages are sent for port instances that are bound prior to the call of this function. Also
     note that turning off binding notifications will not purge 'rtBound' and 'rtUnbound' messages
     that have already been queued.
     @param on_off Set to 0 to turn off binding notifications for the port, and 1 (or any non-zero
     value) to turn them on.
     */
    void bindingNotification( int on_off );

    /** Determine if binding notifications are currently enabled for the port.
     @return 1 if binding notifications have been turned on for this port, 0 otherwise.
     */
    RTS_INLINE int bindingNotificationRequested( void ) const;

    /** Turn on or off type checking when performing send(), invoke() or reply() on the port.
     Calling this function has no effect if the macro RTTYPECHECK_SEND has been set to
     RTTYPECHECK_DONT. Type checking is by default turned on for all ports and can help detect
     problems during development. However, the application will run slightly faster if you turn
     it off.
     @param on_off Set to 0 to turn off type checking for the port, and 1 (or any non-zero value
     to turn it on.
     */
    void sendTypeCheckEnable( int on_off );

    /** Determine if type checking when sending events is enabled.
	 @return 1 if type checking when sending events has been turned on for this port, 0 otherwise.
	 */
    RTS_INLINE int sendTypeCheckEnabled( void ) const;

    /** Check if a particular event would be possible to send on this port. If type checking is
     disabled no check is performed.
     @param signal The id of the event
     @param type The type descriptor of the event data, or 0 if the event does not have any data.
     @return A literal from the RTProtocolDescriptor::Status enum describing if the event could be
     sent on the port or not. "ok" means it would be possible to send the event. "badSignal" means
     the event is not an outgoing (or incoming if the port is conjugated) event in the protocol
     that types the port. "badType" means the type descriptor does not match the type specified
     for the protocol event.
     */
    RTProtocolDescriptor::Status sendCheck( int signal, const RTObject_class * type ) const;

    /** Turn on or off type checking when receiving an event on the port.
	 Calling this function has no effect if the macro RTTYPECHECK_SEND has been set to
	 RTTYPECHECK_DONT. Type checking is by default turned on for all ports and can help detect
	 problems during development. However, the application will run slightly faster if you turn
	 it off.
	 @param on_off Set to 0 to turn off type checking for the port, and 1 (or any non-zero value
	 to turn it on.
	 */
    void receiveTypeCheckEnable( int on_off );

    /** Determine if type checking when receiving events is enabled.
	 @return 1 if type checking on event reception has been turned on for this port, 0 otherwise.
	 */
    RTS_INLINE int receiveTypeCheckEnabled( void ) const;

    /** Check if a particular event would be possible to receive on this port. If type checking is
	 disabled no check is performed.
	 @param signal The id of the event
	 @param type The type descriptor of the event data, or 0 if the event does not have any data.
	 @return A literal from the RTProtocolDescriptor::Status enum describing if the event could be
	 received on the port or not. "ok" means it would be possible to receive the event. "badSignal"
	 means the event is not an incoming (or outgoing if the port is conjugated) event in the
	 protocol that types the port. "badType" means the type descriptor does not match the type
	 specified for the protocol event.
	 */
    RTProtocolDescriptor::Status receiveCheck( int signal, const RTObject_class * type ) const;

protected:
    enum
    {
		BindingNotification 	=  1
		  , RegisteredAsSAP     =  2
		  , SendTypeCheck       =  4
		  , ReceiveTypeCheck    =  8
		  , Resized             = 16
    };

    void                     init( const RTPortDescriptor * );

    RTS_INLINE int           getFlags( void ) const;
    void                     setFlag( int );
    void                     resetFlag( int );

    void                     bindAt( int, RTBindingEnd &, RTController * );
    int                      peerAt( int, RTBindingEnd & );
    void                     unbindAt( int, RTController * );

    void                     notifyInit( RTController * );
    void                     notifyBoundAt( int, RTController * );

    // Derived classes define additional incoming signals starting
    // with ( rtiLast_{BASE} + 1 ) and will define rtiLast_{CLASS}
    // as its last signal.
    enum
    {
    	rtiLast_RTProtocol = 0
    };

private:
    // Represents a port instance
    struct End
    {
		enum State
		{
			Unbound,
			Bound,
			BoundXThread
		};

		State           state;
		RTProtocol    * remote;
		int             remoteIndex;
		const int     * signalMap;

		// these are only called after the priority and type are checked

		int send( RTController         * sender,
				  RTProtocol           * port,
				  int                    index,
				  int                    signal,
				  const void           * data,
				  const RTObject_class * type,
				  int                    prio,
				  bool 					 moveData );

		int invoke( RTMessage            * replyBuffer,
					RTController         * invoker,
					RTProtocol           * port,
					int                    index,
					int                    signal,
					const void           * data,
					const RTObject_class * type,
					bool implicitReply);
    };

    End                    * _ends;
    const RTPortDescriptor * _info;
    int                      _flags;
    RTLayerData            * _layer;

    // unavailable methods
    RTProtocol( const RTProtocol & );
    RTProtocol & operator=( const RTProtocol & );

    friend struct End;
    friend class  RTActor;
    friend class  RTActorProbe;
    friend class  RTDebugger;
    friend class  RTLayerConnector;

    enum CallStyle { Programmatic, Automatic };

    int registerSAP( const char *, CallStyle );
    int deregisterSAP( CallStyle );

    int registerSPP( const char *, CallStyle );
    int deregisterSPP( CallStyle );

    // These classes must have access to the methods below.
    friend struct RTExceptionSignal;
    friend struct RTInSignal;
    friend struct RTOutSignal;
    friend struct RTSymmetricSignal;

    // support for "in" signals

    int purge      ( int signal );
    int purgeAt    ( int signal, int index );

    int recall     ( int signal,            int front );
    int recallAt   ( int signal, int index, int front );

    int recallAll  ( int signal,            int front );
    int recallAllAt( int signal, int index, int front );

public:
    // support for "out" signals
	// These need to be public to support programmatic invoke, reply and send on a general port based on descriptor info

    /** Invoke an event on the port. Normally you should not call this function directly, but
     instead call RTOutsignal::invoke() because it will provide most of the needed parameters for
     you. However, if you need to programatically invoke on a general port based on descriptor info
     this function can be used.
     @param replyBuffer A caller-supplied message object that stores the reply message(s) resulting
     from the invoke. The caller is responsible for allocating and deleting the message(s) when they
     are no longer required. To verify that a returned message is valid call RTMessage::isValid()
     once the invoke returns.
     @param signal The id of the event to invoke
     @param data Data object to pass with the invoke
     @param type Type descriptor describing the data object
     @param implicitReply Whether it should be allowed for the receiver to not make a reply call
     if no data needs to be passed back to the sender.
     @return 1 if the invoke was successful, 0 otherwise.
     */
    int invoke( RTMessage            * replyBuffer,
				int                    signal,
				const void           * data,
				const RTObject_class * type,
				bool implicitReply);

    /** Invoke an event on a replicated port at a specified index. Normally you should not call this
	 function directly, but instead call RTOutsignal::invokeAt() because it will provide most of the
	 needed parameters for you. However, if you need to programatically invoke on a general port
	 based on descriptor info this function can be used.
	 @param index The 0-based index of the port instance within the port on which to perform the
	 invoke (0 <= index < size())
	 @param replyBuffer A caller-supplied message object that stores the reply message(s) resulting
	 from the invoke. The caller is responsible for allocating and deleting the message(s) when they
	 are no longer required. To verify that a returned message is valid call RTMessage::isValid()
	 once the invoke returns.
	 @param signal The id of the event to invoke
	 @param data Data object to pass with the invoke
	 @param type Type descriptor describing the data object
	 @param implicitReply Whether it should be allowed for the receiver to not make a reply call
	 if no data needs to be passed back to the sender.
	 @return 1 if the invoke was successful, 0 otherwise.
	 */
    int invokeAt( int                    index,
				  RTMessage            * replyBuffer,
				  int                    signal,
				  const void           * data,
				  const RTObject_class * type,
				  bool implicitReply);

    /** Make a reply to respond to an invoke. Normally you should not call this function directly,
     but instead call RTOutsignal::reply() because it will provide all of the needed parameters for
     you. However, if you need to programatically reply on a general port based on descriptor info
     this function can be used.
	 @param signal The id of the event to reply with
	 @param data Data object to pass with the reply
	 @param type Type descriptor describing the data object
	 @return 1 if the reply was successful, 0 otherwise.
     */
    int reply( int                    signal,
    		   const void           * data,
			   const RTObject_class * type );

    /** Send an event on the port. Normally you should not call this function directly, but instead
     call RTOutsignal::send() because it will provide most of the needed parameters for you.
     However, if you need to programatically send on a general port based on descriptor info this
     function can be used.
	 @param signal The id of the event to send
	 @param data Data object to pass with the sent event
	 @param type Type descriptor describing the data object
	 @param priority The priority at which the event should be sent. A message priority is
	 interpreted as its relative importance with respect to all other unprocessed messages on a
	 thread. The priority evaluates to one of the defined global priority values.
	 @param moveData Set to true if you want the data object to be moved instead of copied when sending the event
	 @return 1 if the event was successfully sent, 0 otherwise.
     */
    int send( int                    signal,
    		  const void           * data,
			  const RTObject_class * type,
			  int                    priority,
			  bool					 moveData = false);

    /** Send an event on a replicated port at a specified index. Normally you should not call this
	 function directly, but instead call RTOutsignal::sendAt() because it will provide most of the
	 needed parameters for you. However, if you need to programatically send on a general port based
	 on descriptor info this function can be used.
	 @param index The 0-based index of the port instance within the port on which to send the event
	 (0 <= index < size())
	 @param signal The id of the event to send
	 @param data Data object to pass with the sent event
	 @param type Type descriptor describing the data object
	 @param priority The priority at which the event should be sent. A message priority is
	 interpreted as its relative importance with respect to all other unprocessed messages on a
	 thread. The priority evaluates to one of the defined global priority values.
	 @param moveData Set to true if you want the data object to be moved instead of copied when sending the event
	 @return 1 if the event was successfully sent, 0 otherwise.
	 */
    int sendAt( int                    index,
				int                    signal,
				const void           * data,
				const RTObject_class * type,
				int                    priority,
				bool				   moveData = false);

private:
    // support for "exception" signals
    int raise( int                    signal,
			   const void           * data,
			   const RTObject_class * type );
};

#if RTS_INLINES
#include <RTProtocol.inl>
#endif

#endif // __RTProtocol_h__
