/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTSymmetricSignal_h__
#define __RTSymmetricSignal_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTPriority_h__
#include <RTPriority.h>
#endif

#ifndef __RTProtocol_h__
#include <RTProtocol.h>
#endif

// It is common to make temporaries of this struct.
//lint -esym(1901,RTSymmetricSignal)

/** Provides the combined API of both RTOutSignal and RTInSignal for symmetric protocol events
 defined in the model as an incoming and an outgoing event with the same name.
 */
struct RTSymmetricSignal
{
    // general application use

	/** Remove from the defer queue all messages that match the receiver port and the signal id of
	 the protocol event. If the receiver port is replicated messages will be deleted from the defer
	 queue regardless of the port index on which they were received.
	 @return The number of messages deleted from the defer queue.
	 */
    RTS_INLINE int purge      ( void );

    /** Remove from the defer queue all messages that match the receiver port and the signal id of
	 the protocol event. Only the messages that were received on a certain port index will be
	 deleted from the defer queue.
	 @param[in] index Port index (0-based)
	 @return The number of messages deleted from the defer queue.
	 */
    RTS_INLINE int purgeAt    ( int index );

    /** Recall a message from the defer queue and insert it in the controller's message
	 queue (either before or after other queued messages). At most one message will be recalled and
	 it must match the receiver port and the signal id of the protocol event (however, it doesn't
	 matter on which port index it was received).
	 @param[in] front If 0 the recalled message will be inserted at the back of the message queue,
	 otherwise it will be inserted at the front of the message queue.
	 @return The number of messages recalled from the defer queue (i.e. either 1 or 0).
	 */
    RTS_INLINE int recall     ( int front = 0 );

    /** Recall a message from the defer queue and insert it in the controller's message
	 queue (either before or after other queued messages). At most one message will be recalled and
	 it must match the receiver port and the signal id of the protocol event. It must also have
	 been received on a certain port index.
	 @param[in] index Port index (0-based)
	 @param[in] front If 0 the recalled message will be inserted at the back of the message queue,
	 otherwise it will be inserted at the front of the message queue.
	 @return The number of messages recalled from the defer queue (i.e. either 1 or 0).
	 */
    RTS_INLINE int recallAt   ( int index, int front = 0 );

    /** Recall one or many messages from the defer queue and insert them in the controller's message
	 queue (either before or after other queued messages). All messages will be recalled that
	 match the receiver port and the signal id of the protocol event (however, it doesn't
	 matter on which port index they were received).
	 @param[in] front If 0 the recalled messages will be inserted at the back of the message queue,
	 otherwise they will be inserted at the front of the message queue.
	 @return The number of messages recalled from the defer queue.
	 */
    RTS_INLINE int recallAll  ( int front = 0 );

    /** Recall one or many messages from the defer queue and insert them in the controller's message
	 queue (either before or after other queued messages). All messages will be recalled that
	 match the receiver port and the signal id of the protocol event. They must also have
	 been received on a certain port index.
	 @param[in] index Port index (0-based)
	 @param[in] front If 0 the recalled messages will be inserted at the back of the message queue,
	 otherwise they will be inserted at the front of the message queue.
	 @return The number of messages recalled from the defer queue.
	 */
    RTS_INLINE int recallAllAt( int index, int front = 0 );

    /** Invoke a message of the specified protocol event and on the specified port (optionally
	 passing a specified data object with the invoke message).
	 @param replyBuffers A caller-supplied message object that stores the reply message(s) resulting
	 from the invoke. The caller is responsible for allocating and deleting the message(s) when they
	 are no longer required. To verify that a returned message is valid call RTMessage::isValid()
	 once the invoke returns.
	 The callee must make an explicit reply call even if no data needs to be passed back to the
	 caller.
	 @return 1 if the invoke was successful, 0 otherwise.
	 */
    RTS_INLINE int invoke  ( RTMessage * replyBuffers );

    /** Invoke a message of the specified protocol event and on the specified port (optionally
     passing a specified data object with the invoke message). The message is invoked at the
     specified port index.
     @param[in] index Port index (0-based).
     @param replyBuffer A caller-supplied message object that stores the reply message(s) resulting
     from the invoke. The caller is responsible for allocating and deleting the message(s) when they
     are no longer required. To verify that a returned message is valid call RTMessage::isValid()
     once the invoke returns.
	 The callee must make an explicit reply call even if no data needs to be passed back to the
	 caller.
     @return 1 if the invoke was successful, 0 otherwise.
     */
    RTS_INLINE int invokeAt( int index, RTMessage * replyBuffer );

    /** Invoke a message of the specified protocol event and on the specified port (optionally
     passing a specified data object with the invoke message). No reply data can be passed back
     to the caller and the callee does not have to make an explicit reply (it is implicitly made
     when completing the invoke).
     @return 1 if the invoke was successful, 0 otherwise.
     */
    RTS_INLINE int invoke ();

    /** Invoke a message of the specified protocol event and on the specified port (optionally
	 passing a specified data object with the invoke message). The message is invoked at the
     specified port index. No reply data can be passed back to the caller and the callee does not
     have to make an explicit reply (it is implicitly made when completing the invoke).
     @param[in] index Port index (0-based).
	 @return 1 if the invoke was successful, 0 otherwise.
	 */
    RTS_INLINE int invokeAt( int index);

    /** Make a reply to respond to an invoke. The reply message is of the specified protocol event
	 and is made on the specified port (optionally passing a specified data object with the reply
	 message).
	 @return 1 if the reply was successful, 0 otherwise.
	 */
    RTS_INLINE int reply ( void );

    /** Send a message of the specified protocol event and on the specified port (optionally
	 passing a specified data object with the sent message).
	 @param priority The priority at which the message should be sent. A message priority is
	 interpreted as its relative importance with respect to all other unprocessed messages on a
	 thread. The priority evaluates to one of the defined global priority values.
	 @return 1 if the message was successfully sent, 0 otherwise.
	 */
    RTS_INLINE int send ( int priority = General );

    /** Send a message of the specified protocol event and on the specified port (optionally
	 passing a specified data object with the sent message). The message is sent at the
     specified port index.
     @param[in] index Port index (0-based).
	 @param priority The priority at which the message should be sent. A message priority is
	 interpreted as its relative importance with respect to all other unprocessed messages on a
	 thread. The priority evaluates to one of the defined global priority values.
	 @return 1 if the message was successfully sent, 0 otherwise.
	 */
    RTS_INLINE int sendAt ( int index, int priority = General );

    // for use by run-time services library only
    RTS_INLINE RTSymmetricSignal( RTProtocol *, int signal_in, int signal_out );

    /** The port to which the message should be sent. */
    RTProtocol * _port;

    /** The signal id of the incoming protocol event for the received message. */
    int          _in_signal;

    /** The signal id of the outgoing protocol event for the message to send. */
    int          _out_signal;
}; //lint !e1712

#if RTS_INLINES
#include <RTSymmetricSignal.inl>
#endif

#endif // __RTSymmetricSignal_h__
