/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTTimespec_h__
#define __RTTimespec_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTObject_class_h__
#include <RTObject_class.h>
#endif

class RTInterval;
class RTTime;

/** A time value which can represent either an absolute or relative time. For an absolute time the
 representation uses Unix time, i.e. the number of seconds and nanoseconds that have passed since
 1 January 1970.
 */
struct RTTimespec
{
	/** Seconds */
    long tv_sec;
    /** Nanoseconds */
    long tv_nsec;

    /** Construct a zero time value. */
    RTS_INLINE              RTTimespec( void );

    /** Construct a new time value.
     * @param[in] s Number of seconds
     * @param[in] ns Number of nanoseconds
     */
    RTS_INLINE              RTTimespec( long s, long ns);

						    RTTimespec( const RTInterval & ); //lint !e1931
							RTTimespec( const RTTime     & ); //lint !e1931
    RTS_INLINE              RTTimespec( const RTTimespec & );

    /** Normalize the time value so that as many seconds as possible are stored in the tv_sec field
     and remaining nanoseconds in the tv_nsec field. After calling this function it's guaranteed
     that tv_nsec < 1000000000.
     */
    void                    normalize( void );

    RTS_INLINE RTTimespec & operator=( const RTTimespec & );

    RTTimespec            & operator+=( const RTTimespec & );
    RTTimespec            & operator-=( const RTTimespec & );

    /** Get the number of "ticks" which is between this time value and the current time.
     @param[in] nanoseconds_per_tick Defines the length of a "tick" in nanoseconds
     @return The number of "ticks" until this time value is reached. 0 if this time value is before
     the current time.
     */
    long                    ticksFromNow( long nanoseconds_per_tick ) const;

    /** Get the current time.
     @param[out] tspec RTTimespec object to write the current time to.
     */
    static void             getclock( RTTimespec & tspec);

    /** Static constant for a zero time value. */
    static const RTTimespec zero;

    /** Type descriptor for an RTTimespec. */
    static const RTObject_class classData;

#define RTType_RTTimespec ( RTTimespec::classData )
};

RTS_INLINE int operator==( const RTTimespec &, const RTTimespec & );
RTS_INLINE int operator!=( const RTTimespec &, const RTTimespec & );
RTS_INLINE int operator<=( const RTTimespec &, const RTTimespec & );
RTS_INLINE int operator> ( const RTTimespec &, const RTTimespec & );
RTS_INLINE int operator>=( const RTTimespec &, const RTTimespec & );
RTS_INLINE int operator< ( const RTTimespec &, const RTTimespec & );
RTTimespec     operator+ ( const RTTimespec &, const RTTimespec & );
RTTimespec     operator- ( const RTTimespec &, const RTTimespec & );

struct RTTypedValue_RTTimespec
{
    const void           * data;
    const RTObject_class * type;

    RTS_INLINE RTTypedValue_RTTimespec( const RTTimespec     & ); //lint !e1931
    RTS_INLINE RTTypedValue_RTTimespec( const RTTimespec     &,
					const RTObject_class * );
}; //lint !e1712

#if RTS_INLINES
#include <RTTimespec.inl>
#endif

#endif // __RTTimespec_h__
