/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#ifndef __RTTiming_h__
#define __RTTiming_h__ included

#ifdef PRAGMA
#pragma interface
#endif

#ifndef __RTPriority_h__
#include <RTPriority.h>
#endif

#ifndef __RTProtocol_h__
#include <RTProtocol.h>
#endif

#ifndef __RTTimespec_h__
#include <RTTimespec.h>
#endif

class  RTTimerId;
class  RTTimerNode;
struct RTTypedValue;

/** Represents the UML-RT "Timing" protocol */
struct Timing
{
	/** Represents a timer port */
    class Base : public RTProtocol
    {
    public:
		RTS_INLINE Base( void );
		RTS_INLINE ~Base( void );

		static const RTProtocolDescriptor rt_class;

		enum
		{
			rti_timeout = rtiLast_RTProtocol + 1
		};

		/** Return the object that represents the timeout event which is received by this port
		 when the timer has expired.
		 */
		RTS_INLINE RTInSignal timeout( void );

		/** Return the current time. It is recommended, for performance reasons, to instead use
		 RTTimespec::getclock().
		 */
		RTTimespec currentTime( void ) const;

		/** Start a timer which will expire at a particular absolute point in time.
		 @param[in] when The desired absolute time when the timer is to expire.
		 @param[in] data Data object to put in the timeout event. The data is copied by the
		 TargetRTS using the copy function of the type descriptor.
		 @param[in] type Type descriptor for the data object.
		 @param[in] prio The priority at which the timeout event should be sent. A message
		 priority is interpreted as its relative importance with respect to all other unprocessed
		 messages on a thread. The priority evaluates to one of the defined global priority values.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTTimerNode * informAt( const RTTimespec     & when,
					   	   	    const void           * data,
								const RTObject_class * type,
								int                    prio = General );

		/**
		 Start a timer which will expire at a particular absolute point in time.
		 @param[in] when The desired absolute time when the timer is to expire.
		 @param[in] prio The priority at which the timeout event should be sent. A message
		 priority is interpreted as its relative importance with respect to all other unprocessed
		 messages on a thread. The priority evaluates to one of the defined global priority values.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTS_INLINE RTTimerNode * informAt( const RTTimespec   & when,
				   	   	   	   	   	   	   int                  prio = General );

		/** Start a timer which will expire at a particular absolute point in time.
		 @param[in] when The desired absolute time when the timer is to expire.
		 @param[in] data Data object to put in the timeout event. The data is copied by the
		 TargetRTS using the copy function of the type descriptor for the data object.
		 @param[in] prio The priority at which the timeout event should be sent. A message
		 priority is interpreted as its relative importance with respect to all other unprocessed
		 messages on a thread. The priority evaluates to one of the defined global priority values.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTS_INLINE RTTimerNode * informAt( const RTTimespec   & when,
				   	   	   	   	   	   	   const RTDataObject & data,
										   int                  prio = General );

		/** Start a timer which will expire at a particular absolute point in time.
		 @param[in] when The desired absolute time when the timer is to expire.
		 @param[in] info An RTTypedValue object which encapsules the data object to put in the
		 timeout event and its type descriptor. The data is copied by the TargetRTS using the copy
		 function of the type descriptor.
		 @param[in] prio The priority at which the timeout event should be sent. A message
		 priority is interpreted as its relative importance with respect to all other unprocessed
		 messages on a thread. The priority evaluates to one of the defined global priority values.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTS_INLINE RTTimerNode * informAt( const RTTimespec   & when,
					   	   	   	   	   	   const RTTypedValue & info,
										   int                  prio = General );

		/** Start a timer which will expire after a specified time interval.
		 @param[in] delta The desired time interval (expressed as a time relative from now) when the
		 timer is to expire. If the interval is less than or equal to the current time or equal to
		 zero, the timer will expire immediately.
		 @param[in] data Data object to put in the timeout event. The data is copied by the
		 TargetRTS using the copy function of the type descriptor.
		 @param[in] type Type descriptor for the data object.
		 @param[in] prio The priority at which the timeout event should be sent. A message
		 priority is interpreted as its relative importance with respect to all other unprocessed
		 messages on a thread. The priority evaluates to one of the defined global priority values.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTS_INLINE RTTimerNode * informIn( const RTTimespec     & delta,
										   const void           * data,
										   const RTObject_class * type,
										   int                    prio = General );

		/** Start a timer which will expire after a specified time interval.
		 @param[in] delta The desired time interval (expressed as a time relative from now) when the
		 timer is to expire. If the interval is less than or equal to the current time or equal to
		 zero, the timer will expire immediately.
		 @param[in] prio The priority at which the timeout event should be sent. A message
		 priority is interpreted as its relative importance with respect to all other unprocessed
		 messages on a thread. The priority evaluates to one of the defined global priority values.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTS_INLINE RTTimerNode * informIn( const RTTimespec   & delta,
				   	   	   	   	   	   	   int                  prio = General );

		/** Start a timer which will expire after a specified time interval.
		 @param[in] delta The desired time interval (expressed as a time relative from now) when the
		 timer is to expire. If the interval is less than or equal to the current time or equal to
		 zero, the timer will expire immediately.
		 @param[in] data Data object to put in the timeout event. The data is copied by the
		 TargetRTS using the copy function of the type descriptor for the data object.
		 @param[in] prio The priority at which the timeout event should be sent. A message
		 priority is interpreted as its relative importance with respect to all other unprocessed
		 messages on a thread. The priority evaluates to one of the defined global priority values.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTS_INLINE RTTimerNode * informIn( const RTTimespec   & delta,
				   	   	   	   	   	   	   const RTDataObject & data,
										   int                  prio = General );

		/** Start a timer which will expire after a specified time interval.
		 @param[in] delta The desired time interval (expressed as a time relative from now) when the
		 timer is to expire. If the interval is less than or equal to the current time or equal to
		 zero, the timer will expire immediately.
		 @param[in] info An RTTypedValue object which encapsules the data object to put in the
		 timeout event and its type descriptor. The data is copied by the TargetRTS using the copy
		 function of the type descriptor.
		 @param[in] prio The priority at which the timeout event should be sent. A message
		 priority is interpreted as its relative importance with respect to all other unprocessed
		 messages on a thread. The priority evaluates to one of the defined global priority values.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTS_INLINE RTTimerNode * informIn( const RTTimespec   & delta,
				   	   	   	   	   	   	   const RTTypedValue & info,
										   int                  prio = General );

#if RTS_COMPATIBLE <= 640
		RTTimerNode    * informIn( unsigned long        centiseconds,
					   int                  prio = General );

		RTTimerNode    * informIn( unsigned long        centiseconds,
					   const RTDataObject & data,
					   int                  prio = General );

		RTTimerNode    * informIn( unsigned long        centiseconds,
					   const RTTypedValue & info,
					   int                  prio = General );
#endif // RTS_COMPATIBLE <= 640

		/** Start a periodic timer which will expire at regular intervals.
		 @param[in] delta The desired time interval specifying how frequently the timer should
		 expire. If the interval is less than or equal to the current time or equal to zero, the
		 timer will expire immediately.
		 @param[in] data Data object to put in the timeout event. The data is copied by the
		 TargetRTS using the copy function of the type descriptor.
		 @param[in] type Type descriptor for the data object.
		 @param[in] prio The priority at which the timeout event should be sent. A message
		 priority is interpreted as its relative importance with respect to all other unprocessed
		 messages on a thread. The priority evaluates to one of the defined global priority values.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTS_INLINE RTTimerNode * informEvery( const RTTimespec     & delta,
											  const void           * data,
											  const RTObject_class * type,
											  int                    prio = General );

		/** Start a periodic timer which will expire at regular intervals.
		 @param[in] delta The desired time interval specifying how frequently the timer should
		 expire. If the interval is less than or equal to the current time or equal to zero, the
		 timer will expire immediately.
		 @param[in] data Data object to put in the timeout event. The data is copied by the
		 TargetRTS using the copy function of the type descriptor for the data object.
		 @param[in] prio The priority at which the timeout event should be sent. A message
		 priority is interpreted as its relative importance with respect to all other unprocessed
		 messages on a thread. The priority evaluates to one of the defined global priority values.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTS_INLINE RTTimerNode * informEvery( const RTTimespec   & delta,
				      	  	  	  	  	  	  const RTDataObject & data,
											  int                  prio = General );

		/** Start a periodic timer which will expire at regular intervals.
		 @param[in] delta The desired time interval specifying how frequently the timer should
		 expire. If the interval is less than or equal to the current time or equal to zero, the
		 timer will expire immediately.
		 @param[in] info An RTTypedValue object which encapsules the data object to put in the
		 timeout event and its type descriptor. The data is copied by the TargetRTS using the copy
		 function of the type descriptor.
		 @param[in] prio The priority at which the timeout event should be sent. A message
		 priority is interpreted as its relative importance with respect to all other unprocessed
		 messages on a thread. The priority evaluates to one of the defined global priority values.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTS_INLINE RTTimerNode * informEvery( const RTTimespec   & delta,
				      	  	  	  	  	  	  const RTTypedValue & info,
											  int                  prio = General );

		/** Start a periodic timer which will expire at regular intervals.
		 @param[in] delta The desired time interval specifying how frequently the timer should
		 expire. If the interval is less than or equal to the current time or equal to zero, the
		 timer will expire immediately.
		 @return A timer handle which can be used to construct a timer identifier represented by an
		 RTTimerId. This object can be used to cancel the timer prior to its expiry. NULL is
		 returned in case the timer request fails.
		 */
		RTS_INLINE RTTimerNode * informEvery( const RTTimespec & delta );

		/** Get the number of timeout intervals that passed before the timeout of a periodic timer
		 was actually processed. Calling this function is a way to measure the accuracy of a
		 periodic timer. The precision of a periodic timer depends on the underlying operating
		 system. If you set a periodic timer with a timeout interval that is small, you may miss
		 certain timeouts simply because the underlying operating system cannot process the timeouts
		 fast enough. This function returns how many timeout intervals that have been missed since
		 the last time a periodic timer expired. NOTE: Setting a timeout frequency higher than the
		 granularity of the operating system means that the memory usage will grow since timeout
		 messages then are produced at a higher pace than what they are consumed. If this goes on
		 for long enough the application will run out of memory.
		 @param[in] id Timer identifier for a periodic timer to measure the accuracy of.
		 @return A measurement of the number of missed timeout intervals since the last time the
		 timer expired. If the timer never has expired 0 is returned. If no interval is missed (i.e.
		 the timer expired at the specified time) 1 is returned. The higher number returned, the
		 less accurate is the timer.
		 */
		int timeouts( const RTTimerId & id );

		/** Cancel a timer. The function ensures that the timeout event of the specified timer will
		 not be delivered to this timer port. This is true even if the timer already have expired
		 at the time when this function is called, but the timeout event is waiting to be dispatched.
		 @param[in] id Timer identifier for the timer to be cancelled.
		 @return 1 if a pending timeout request was cancelled, 0 otherwise.
		 */
		int cancelTimer( RTTimerId & id );

		/** Begin adjustment of system time. If ever the real-time clock needs to be adjusted it has
		 to be done between calls to adjustTimeBegin() and adjustTimeEnd() so that the run-time
		 system can be aware of it and make any necessary adjustments to its internal data
		 structures so that relative timeouts (requested via informIn() or informEvery() calls)
		 are not delayed nor advanced.
		 */
		void adjustTimeBegin( void );

		/** End adjustment of system time. If ever the real-time clock needs to be adjusted it has
		 to be done between calls to adjustTimeBegin() and adjustTimeEnd() so that the run-time
		 system can be aware of it and make any necessary adjustments to its internal data
		 structures so that relative timeouts (requested via informIn() or informEvery() calls)
		 are not delayed nor advanced.
		 @param[in] delta Specifies how the real-time clock was changed
		 */
		void adjustTimeEnd( const RTTimespec & delta );

    protected:
		RTTimerNode * informIn( const RTTimespec     & when,
				   	   	   	   	const RTTimespec     & repeatInterval,
								const void           * data,
								const RTObject_class * type,
								int                    prio );

		enum
		{
			rtiLast_Timing = rti_timeout
		};

    private:
		friend class RTTimerActor;
		friend class RTTimerNode;

		RTTimerNode * outstanding;

		// unavailable methods
		Base( const Base & );
		Base & operator=( const Base & );
    };
};

#if RTS_INLINES
#include <RTTiming.inl>
#endif

#endif // __RTTiming_h__
