/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#include <RTActor.h>
#include <RTActorClass.h>
#include <RTActorFactoryInterface.h>
#include <RTActorRef.h>
#include <RTComponentDescriptor.h>
#include <RTDebugger.h>

RTActor * RTActorRef::incarnate2( const RTActorClass * actorClass,
		RTController       		* rts,
		int                  	  index,
		RTActorFactoryInterface	* factory )
{
	RTActor      * myOwner    = getOwner();
	RTController * controller = myOwner->context();

	controller->trace( RTController::FrameIncarnate, this, &index );

	if( _info->kind == RTComponentDescriptor::Imported )
	{
		controller->setError( RTController::imported,
				RTController::FrameIncarnate, this, &index );
		return nullptr;
	}

	RTController::Error error = index < 0 ? RTController::refFull
			: RTController::badIndex;

	if( reserve( index ) == 0 )
	{
		controller->setError( error, RTController::FrameIncarnate, this, &index );
		return nullptr;
	}

	if( rts == nullptr )
	{
		rts = myOwner->defaultController( *this, index );

		if( rts == nullptr )
			rts = controller;
	}

	// We need either a capsule factory or a capsule type descriptor with a create function to be able to
	// create an instance of the capsule
	if( actorClass == nullptr || actorClass == &EmptyActorClass ||
			(actorClass->create == nullptr && !factory) )
	{
		actorClass = myOwner->defaultClass( *this, index );

		if( actorClass == nullptr || (actorClass->create == nullptr && !factory) )
		{
			controller->setError( RTController::badClass,
					RTController::FrameIncarnate, this, &index );
			unreserve( index );
			return nullptr;
		}
	}

	error = controller->canPlugIn( actorClass, _info );

	if( error != RTController::ok )
	{
		controller->setError( error, RTController::FrameIncarnate, this, &index );

#if RTTYPECHECK_PROTOCOL == RTTYPECHECK_WARN
		if( error != RTController::protocolType )
#endif
		{
			unreserve( index );
			return nullptr;
		}
	}

	// Now create the capsule instance, either using the provided factory, or the create function of the capsule type descriptor
	RTActor * actor = factory ?
			factory->create( rts, this, index )
			: (*actorClass->create)( rts, this );

	if( actor == nullptr )
	{
		controller->setError( RTController::badClass,
				RTController::FrameIncarnate, this, &index );
		unreserve( index );
		return nullptr;
	}

	actor->_preinitialize( index, _info );

#if USE_THREADS && RTFRAME_THREAD_SAFE
	lock();
#endif

	add( actor, index );

	if( bind_ports( actor, index, nullptr ) == 0 )
	{
#if USE_THREADS && RTFRAME_THREAD_SAFE
		unlock();
#endif
		controller->setError( RTController::badClass,
				RTController::FrameIncarnate, this, &index );
		remove( index );
		if (factory)
			factory->destroy(actor);
		else
			delete actor;
		return nullptr;
	}

#if USE_THREADS && RTFRAME_THREAD_SAFE
	unlock();
#endif

	actor->_incarnateAllFixed();

#if OBSERVABLE
	controller->debugger()->actorIncarnated( actor );
#endif

	return actor;
}
