/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#include <RTActorRef.h>
#include <RTActorSlot.h>
#include <RTComponentDescriptor.h>
#include <RTDiag.h>
#include <RTImportList.h>

int RTActorRef::reserve( int & index )
{
	int replication = _info->replication;

	if( index >= replication )
		return 0; // index out of range

	int status = 1;

#if USE_THREADS
	lock();
#endif

	switch( _info->kind )
	{
	case RTComponentDescriptor::Imported:
	{
		RTImportList * entry;

		if( replication == 1 )
		{
			// We don't have a free list, try to allocate the only entry.
			entry = static_cast<RTImportList *>(_contents);
			if( entry->used.reference == this )
			{
				status = 0; // already allocated
				break;
			}
			index = 0;
		}
		else
		{
			RTImportList::Unused * head  = static_cast<RTImportList::Unused *>(_contents);
			RTImportList         * first = static_cast<RTImportList *>(static_cast<void *>(&head[ 1 ]));

			if( index < 0 )
			{
				if( head == reinterpret_cast<RTImportList::Unused *>(head->next) )
				{
					status = 0; // full
					break;
				}

				entry = head->next;
				index = static_cast<int>(entry - first);
			}
			else
			{
				entry = &first[ index ];

				if( entry->used.reference == this )
				{
					status = 0; // already allocated
					break;
				}
			}

			entry->unused.prev->unused.next = entry->unused.next;
			entry->unused.next->unused.prev = entry->unused.prev;
		}

		entry->used.reference = this;
		break;
	}

	case RTComponentDescriptor::Optional:
		if( replication == 1 )
		{
			if( _contents != nullptr )
				status = 0; // in use
			// We don't have a free list, allocate the only entry.
			index = 0;
		}
		else if( RTSmallOptional::supports( replication ) != 0 )
		{
			RTSmallOptional * head  = static_cast<RTSmallOptional *>(_contents);
			RTSmallOptional * entry;

			if( index < 0 )
			{
				int iReserved = head->next();

				if( iReserved == 0 )
				{
					status = 0; // full
					break;
				}

				index = iReserved - 1;
				entry = &head[ iReserved ];
			}
			else
			{
				entry = &head[ index + 1 ];
				if( entry->inuse() != 0 )
				{
					status = 0; // index already allocated
					break;
				}
			}

			int               iPrev = entry->prev();
			int               iNext = entry->next();
			RTSmallOptional * prev  = &head[ iPrev ];
			RTSmallOptional * next  = &head[ iNext ];

			prev->next( iNext );
			next->prev( iPrev );

			entry->_actor = nullptr;
		}
		else
		{
			//lint -save -e661
			RTLargeOptional * head  = static_cast<RTLargeOptional *>(_contents);
			RTLargeOptional * entry;

			if( index < 0 )
			{
				entry = head->next;

				if( entry == head )
				{
					status = 0; // full
					break;
				}

				index = static_cast<int>( entry - head ) - 1;
			}
			else
			{
				entry = &head[ index + 1 ];

				if( entry->next == nullptr )
				{
					status = 0; // index already allocated
					break;
				}
			}

			entry->prev->next = entry->next;
			entry->next->prev = entry->prev;

			entry->next  = nullptr;
			entry->actor = nullptr;
			//lint -restore
		}
		break;

	case RTComponentDescriptor::Fixed:
		if( index < 0 )
		{
			// We are always specific with fixed components,
			// this must be the application doing something illegal.
			status = 0;
		}
		else if( replication == 1 )
		{
			if( _contents != nullptr )
				status = 0; // in use
		}
		else if( (static_cast<RTActor **>(_contents))[ index ] != nullptr )
		{
			status = 0; // in use
		}
		break;

	default:
		status = 0;
		break;
	}

#if USE_THREADS
	unlock();
#endif

	return status;
}
