/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#include <RTCachedString.h>
#include <RTDiag.h>
#include <RTDictionary.h>
#include <RTMemoryUtil.h>

unsigned                RTCachedString::_size;
unsigned                RTCachedString::_used;
RTCachedString::Block * RTCachedString::_blocks;
RTCachedString::Entry** RTCachedString::_contents = nullptr;

static RTCachedString::Block * new_Block( RTCachedString::Block * next )
{
	RTCachedString::Block * block = new RTCachedString::Block;
	RTCachedString::Entry * entry;
	int                     count;

	block->next = next;

	for( entry = &block->entry[ 0 ], count = RTCachedString::Block::Size;
			--count >= 0;
			++entry )
	{
		entry->value     = nullptr;
		entry->use_count = 0U;
	}

	return block;
}

RTCachedString::Entry * RTCachedString::add( const char * value )
{
	if( value == nullptr )
		return nullptr;

	if( _contents == nullptr )
	{
		Block * block = new_Block( nullptr );
		Entry * entry;
		int     index;

		_size     = Block::Size;
		_used     = 0U;
		_contents = new Entry * [ Block::Size ];
		_blocks   = block;

		for( entry = &block->entry[ 0 ], index = 0;
				index < Block::Size;
				++entry,                    ++index )
		{
			_contents[ index ] = entry;
		}
	}

	for(;;) // loop back (at most once) if we need to expand the table
	{
		unsigned   index   = RTDictionary::hash( value, _size );
		unsigned   tests   = 0U;
		Entry    * deleted = nullptr;
		Entry    * entry;

		for(;;)
		{
			entry = _contents[ index ];

			if( entry->value == nullptr )
			{
				// this slot has never been used
				if( deleted != nullptr )
					entry = deleted; // use a deleted entry, if we saw one
				break;
			}

			if( entry->use_count == 0U )
			{
				if( deleted == nullptr )
					deleted = entry;
			}
			else if( RTMemoryUtil::strcmp( entry->value, value ) == 0 )
			{
				// the value already exists
				entry->use_count += 1U;
				return entry;
			}

			if( ++tests == _size )
			{
				// we looked everywhere; use the first deleted entry, if any
				entry = deleted;
				break;
			}

			if( ++index == _size )
				index = 0U;
		}

		if( entry != nullptr )
		{
			entry->value     = RTMemoryUtil::strdup( value );
			entry->use_count = 1U;

			++_used;

			return entry;
		}

		// the table needs to be expanded

		unsigned    old_size     = _size;
		Entry    ** old_contents = _contents;

		_size   <<= 1U;
		_contents = new Entry * [ _size ];

		for( index = 0U; index < _size; ++index )
			_contents[ index ] = nullptr;

		for( unsigned old_index = 0U; old_index < old_size; ++old_index )
		{
			entry = old_contents[ old_index ];
			index = RTDictionary::hash( entry->value, _size );

			// search for an empty slot
			while( _contents[ index ] != nullptr )
				if( ++index == _size )
					index = 0U;

			_contents[ index ] = entry;
		}

		delete [] old_contents;

		// allocate blocks for the half that is still empty
		{
			unsigned count = 0U;

			for( index = 0U; index < _size; ++index )
			{
				if( _contents[ index ] == nullptr )
				{
					if( count == 0U )
					{
						_blocks = new_Block( _blocks );
						entry = &_blocks->entry[ 0 ];
						count = Block::Size;
					}
					_contents[ index ] = entry++; //lint !e794
					--count;
				}
			}

			RTAssert( count == 0U );
		}
	}
}
