/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#include <RTActor.h>
#include <RTActorFactoryInterface.h>
#include <RTActorRef.h>
#include <RTController.h>
#include <RTDebugger.h>
#include <RTDiag.h>
#include <RTMessage.h>
#include <RTQFilter.h>
#include <RTSpecialSignals.h>

class RTController_destroy_filter : public RTQFilter
{
public:
	RTController_destroy_filter( RTActor      * _actor,
			RTController * _rts )
: RTQFilter()
, actor( _actor )
, rts  ( _rts   )
{
}
	virtual ~RTController_destroy_filter( void );

	virtual int  select( RTMessage & ) override;
	virtual void accept( RTMessage & ) override;

private:
	RTActor      * actor;
	RTController * rts;

	// unavailable methods
	RTController_destroy_filter( const RTController_destroy_filter & );
	RTController_destroy_filter & operator=( const RTController_destroy_filter & );
}; //lint !e1712

int RTController_destroy_filter::select( RTMessage & msg )
{
	return actor == msg.receiver();
}

void RTController_destroy_filter::accept( RTMessage & msg )
{
	rts->freeMsg( &msg );
}

RTController_destroy_filter::~RTController_destroy_filter( void )
{
	actor = nullptr;
	rts   = nullptr;
}

void RTController::destroy( RTActor * actor, RTActor * owner )
{
	(void)owner; // it is used (by the threaded variant)

#if USE_THREADS
	RTController * target = actor->context();

	if( target != this )
	{
		if( send( actor, RTDestroySignal, &owner, nullptr, Synchronous ) != 0 )
		{
			while( owner != nullptr )
				awaitResponse();
		}
		return;
	}
#endif

#if OBSERVABLE
	debugger()->actorDestroyed( actor );
#endif

	actor->_predestroy();

	RTController_destroy_filter filter( actor, this );

	filterMessages( actor, filter );

	// If the capsule instance is in a capsule part with a capsule factory defined, let the factory
	// destroy it. Otherwise just delete the capsule instance as usual.
	RTActorRef* ref = actor->getReference();
	RTActorFactoryInterface* factory = (ref && ref->getInfo()) ? ref->getInfo()->factory : nullptr;
	if (factory)
		factory->destroy(actor);
	else
		delete actor;
}

//lint -save -e1738
RTController_destroy_filter::
RTController_destroy_filter( const RTController_destroy_filter & )
: RTQFilter()
, actor( nullptr )
, rts  ( nullptr )
{
	RTDiag::panic( "Unavailable method called:\n"
			"    'RTController_destroy_filter( const RTController_destroy_filter & )'" );
}
//lint -restore

RTController_destroy_filter &
RTController_destroy_filter::operator=( const RTController_destroy_filter & )
{
	RTDiag::panic( "Unavailable method called:\n"
			"    'RTController_destroy_filter::operator=( const RTController_destroy_filter & )'" );

	//lint -unreachable
	return *this;
} //lint !e1529
