/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#include <RTActorRef.h>
#include <RTBindingEnd.h>
#include <RTController.h>
#include <RTDebugger.h>
#include <RTDiag.h>
#include <RTDiagStream.h>
#include <RTProtocol.h>

void RTDebugger::trace( RTController::Primitive   primitive,
		const char              * errorDescription,
		const void              * info,
		const void              * detail )
{
	enum InfoType
	{
		NoInfo,
		ActorRefAndIndex,
		BindingEnd,
		Message,
		PortAndIndex,
		PortAndName,
		PortAndSize
	};

	const char                * prim_name;
	RTDebugger::TraceCategory   prim_type;
	InfoType                    info_type;

	switch( primitive )
	{
	case RTController::CommDefer:
		prim_name = "defer";
		info_type = Message;
		prim_type = TraceComm;
		break;
	case RTController::CommDeliver:
		prim_name = "deliver";
		info_type = Message;
		prim_type = TraceComm;
		break;
	case RTController::CommInvoke:
		prim_name = "invoke";
		info_type = PortAndIndex;
		prim_type = TraceComm;
		break;
	case RTController::CommPurge:
		prim_name = "purge";
		info_type = PortAndIndex;
		prim_type = TraceComm;
		break;
	case RTController::CommRecall:
		prim_name = "recall";
		info_type = PortAndIndex;
		prim_type = TraceComm;
		break;
	case RTController::CommRecallAll:
		prim_name = "recallAll";
		info_type = PortAndIndex;
		prim_type = TraceComm;
		break;
	case RTController::CommReply:
		prim_name = "reply";
		info_type = PortAndIndex;
		prim_type = TraceComm;
		break;
	case RTController::CommSend:
		prim_name = "send";
		info_type = PortAndIndex;
		prim_type = TraceComm;
		break;

	case RTController::ExceptionRaise:
		prim_name = "raise";
		info_type = PortAndIndex;
		prim_type = TraceException;
		break;

	case RTController::FrameDeport:
		prim_name = "deport";
		info_type = ActorRefAndIndex;
		prim_type = TraceFrame;
		break;
	case RTController::FrameDestroy:
		prim_name = "destroy";
		info_type = ActorRefAndIndex;
		prim_type = TraceFrame;
		break;
	case RTController::FrameImport:
		prim_name = "import";
		info_type = ActorRefAndIndex;
		prim_type = TraceFrame;
		break;
	case RTController::FrameIncarnate:
		prim_name = "incarnate";
		info_type = ActorRefAndIndex;
		prim_type = TraceFrame;
		break;

	case RTController::LayerBind:
		prim_name = "bind";
		info_type = BindingEnd;
		prim_type = TraceLayer;
		break;
	case RTController::LayerDeregisterSAP:
		prim_name = "deregisterSAP";
		info_type = PortAndName;
		prim_type = TraceLayer;
		break;
	case RTController::LayerRegisterSAP:
		prim_name = "registerSAP";
		info_type = PortAndName;
		prim_type = TraceLayer;
		break;
	case RTController::LayerDeregisterSPP:
		prim_name = "deregisterSPP";
		info_type = PortAndName;
		prim_type = TraceLayer;
		break;
	case RTController::LayerRegisterSPP:
		prim_name = "registerSPP";
		info_type = PortAndName;
		prim_type = TraceLayer;
		break;
	case RTController::LayerResize:
	case RTController::LayerResizeSAP:
	case RTController::LayerResizeSPP:
		prim_name = "resize";
		info_type = PortAndSize;
		prim_type = TraceLayer;
		break;

	case RTController::TimerCancel:
		prim_name = "cancel";
		info_type = PortAndIndex;
		prim_type = TraceTimer;
		break;
	case RTController::TimerInform:
		prim_name = "inform";
		info_type = PortAndIndex;
		prim_type = TraceTimer;
		break;
	case RTController::SystemService:
		prim_name = "system";
		info_type = NoInfo;
		prim_type = TraceSystem;
		break;

	default:
		RTDiag::panic( "trace: unknown primitive" );
		//lint -unreachable
		return;
	}

	switch( traceLevel[ prim_type ] )
	{
	case DEBUG_NONE:
		return;
	case DEBUG_TERSE:
		if( errorDescription == nullptr )
			return;
		break;
	case DEBUG_VERBOSE:
	default:
		break;
	}

	RTDiagStream & out = RTDiag::out();

	out.lock() << prim_name << "(";

	switch( info_type )
	{
	case ActorRefAndIndex:
	{
		const RTActorRef * ref = static_cast<const RTActorRef *>(info);

		if( ref == nullptr )
		{
			out << "<nil>";
			break;
		}

		printRef( ref->getOwner(), ref->getName() );

		if( detail == nullptr )
			break;

		out << "[" << * static_cast<const int *>(detail) << "]";
		break;
	}

	case BindingEnd:
	{
		const RTBindingEnd * binding;

		binding = static_cast<const RTBindingEnd *>(info);
		printPort( binding->port, binding->index );

		if( detail == nullptr )
			break;

		out << " to ";
		binding = static_cast<const RTBindingEnd *>(detail);
		printPort( binding->port, binding->index );
		break;
	}

	case PortAndIndex:
	case PortAndName:
	case PortAndSize:
	{
		const RTProtocol * port = static_cast<const RTProtocol *>(info);

		if( port == nullptr )
		{
			out << "<nil>";
			break;
		}

		printRef( port->getOwner(), port->getName() );

		if( detail == nullptr )
			break;

		switch( info_type )
		{
		case PortAndIndex:
			out << "["    << * static_cast<const int *>(detail) << "]";
			break;
		case PortAndName:
			out << " as " <<  static_cast<const char *>(detail);
			break;
		case PortAndSize:
			out << "("    << * static_cast<const int *>(detail) << ")";
			break;
		default:
			break;
		}
		break;
	}

	case Message:
		// print the message below
		break;

	case NoInfo:
	default:
		break;
	}

	if( errorDescription == nullptr )
		out << ")\n";
	else
		out << "): " << errorDescription << "\n";

	if( info_type == Message )
		printMessage( static_cast<const RTMessage *>(info) );

	out.unlock();
}
