/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#include <RTActor.h>
#include <RTActorProbe.h>
#include <RTController.h>
#include <RTDebugger.h>
#include <RTDiag.h>
#include <RTMessage.h>
#include <RTPortDescriptor.h>
#include <RTProtocol.h>
#include <RTProtocolDescriptor.h>
#include <RTSpecialPorts.h>
#include <RTSpecialSignals.h>
#include <RTTimerNode.h>

RTMessage::DeliverResult RTMessage::deliver( void )
{
	RTActor      * destination = _receiver;
	RTController * destContext = destination->context();
	int            id;

	if( _port == nullptr )
	{
		id = RTControlPort;
		switch( signal )
		{
		case RTInitSignal:
			// some other thread asked to init an actor
			destination->msg = this;

#if LOG_MESSAGE
			destination->logMsg();
#endif

			destination->initializeAll();
			destination->processHistory();

#if OBSERVABLE
			if( destination->probe != nullptr )
				destination->probe->evtChain( 0 );
#endif

			// message was delivered, now clear msg in destination
			destination->msg = nullptr;

			return Accepted;

		case RTDestroySignal:
			// some other thread asked to destroy an actor
		{
			RTActor * owner = * static_cast<RTActor **>(data); //lint !e413 !e1402

			RTAssert( owner != nullptr );

			destContext->destroy( destination, owner );
			if( destContext->send( owner, RTAckSignal, data, nullptr, Synchronous ) == 0 )
				RTDiag::panic( "Can't reply to destroy signal" );
		}
		return Accepted;

		case RTAckSignal:
			// This is an acknowledgement: the request sent by 'receiver'
			// or on its behalf has been completed. Clearing the pointer
			// addressed by data will accept this acknowledgement
			* static_cast<void **>(data) = nullptr; //lint !e413
			return Accepted;

		case RTTimeoutSignal:
			return RTTimerNode::deliver( this );

		case RTPendingTimeoutSignal:
			// This message should not have been queued!
			RTDiag::panic( "Queued pending timeout" );
			//lint -unreachable
			return Accepted;

#if OBSERVABLE
		case RTProbeSignal:
			// probe wants attention, possibly ESV read/write
			destContext->debugger()->probeAttention( destination );
			return Accepted;
#endif

		case RTInterruptSignal:
			break;

		case RTStartSignal:
		case RTStopSignal:
			RTDiag::panic( "Remote start and stop not yet implemented" );
			//lint -unreachable
			return Accepted;

		case RTWakeupSignal:
			// This was just for waking up the message-receiving-thread.
			// The thread has already been awaken. Now is the time to
			// gracefully handle this signal.
			return Accepted;

		default:
			RTDiag::panic( "There is no such control message!" );
			//lint -unreachable
			return Accepted;
		}
	}
	else
		id = _port->getId();

#if SUSPEND
	if( destination->isSuspended() )
	{
		next = destination->suspendMsgs;
		destination->suspendMsgs = this;
		return Deferred;
	}
#endif

	destination->msg = this;

#if OBSERVABLE
	RTActorProbe * probe = destination->probe;

	if( probe != nullptr )
	{
		probe->evtMessageIn( this );

		if( destContext->getReplyBuffer() != nullptr )
			probe->evtMessageOutIn( this, RTEventInfo::SendInvoke );
		else
			probe->evtMessageOutIn( this, RTEventInfo::SendNormal );
	}
#endif

#if LOG_MESSAGE
	destination->logMsg();
#endif

	RTMessage::DeliverResult result = Accepted;

#if RTTYPECHECK_RECEIVE != RTTYPECHECK_DONT
	if( _port == nullptr )
	{
	}
	else switch( _port->receiveCheck( signal, _type ) )
	{
	case RTProtocolDescriptor::badType:
		destContext->setError( RTController::badClass,
				RTController::CommDeliver,
				this, nullptr );
#if RTTYPECHECK_RECEIVE == RTTYPECHECK_FAIL
		result = Rejected;
#endif
		break;

	case RTProtocolDescriptor::badSignal:
		destContext->setError( RTController::badSignal,
				RTController::CommDeliver,
				this, nullptr );
#if RTTYPECHECK_RECEIVE == RTTYPECHECK_FAIL
		result = Rejected;
#endif
		break;

	case RTProtocolDescriptor::ok:
	default:
		break;
	}
#endif // RTTYPECHECK_RECEIVE != RTTYPECHECK_DONT

#if RTTYPECHECK_RECEIVE == RTTYPECHECK_FAIL
	if( result == Accepted )
#endif
	{
		destination->rtsBehavior( signal, id );
#if OBSERVABLE
		if( destination->probe )
			destination->probe->reportFieldValues( );
#endif
	}

#if OBSERVABLE
	if( probe != nullptr )
		probe->evtChain( 0 );
#endif

	// message was delivered, now clear msg in destination
	destination->msg = nullptr;

	if( isDeferred() != 0 )
	{
#if DEFER_IN_ACTOR
		destination->deferredQ.nqTail( this );
#else
		destContext->deferredQ.nqTail( this );
#endif
		result = Deferred;
	}

	return result;
}
