/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#include <RTDaemon.h>
#include <RTDaemonInfo.h>
#include <RTEventInfo.h>
#include <RTMessage.h>
#include <RTObserver.h>
#include <RTPortDescriptor.h>
#include <RTProtocol.h>

void RTObserver::dispatchEvent( RTEventInfo & eventInfo )
{
	int shouldHalt = 0;

	if( eventInfo.eventKind  == RTEventInfo::EventControl &&
			eventInfo.targetMode == RTEventInfo::Running      )
	{
		resumeToRun = 1;
	}

	for( int daemonId = 0; daemonId < daemons.size(); ++daemonId )
	{
		// Only treat daemonId's that are < 0 or match
		if( eventInfo.daemonId >= 0 && eventInfo.daemonId != daemonId )
			continue;

		RTDaemon * theDaemon = static_cast<RTDaemon *>(daemons[ daemonId ]);

		if( theDaemon == nullptr )
			continue;

		// update the daemon's settings
		if( eventInfo.eventKind == RTEventInfo::EventPresence )
			theDaemon->setInfo( eventInfo );

		if( theDaemon->setting( eventInfo.eventKind ) == 0 )
			continue;

		if( eventInfo.eventKind == RTEventInfo::EventControl )
		{
			shouldHalt += evtRtsStatus( daemonId, eventInfo );
			continue;
		}

		RTDaemonInfo * daemonInfo = theDaemon->eventMatches( eventInfo );

		if( daemonInfo == nullptr )
			continue;

		if( superDaemon->shouldReport() == 0 )
			continue;

		if( theDaemon->shouldReport() == 0 )
		{
			// If we're halting, send a halted message to the daemon.
			if( theDaemon->shouldHalt() != 0 )
			{
				eventInfo.targetMode = RTEventInfo::Halted;
				(void)evtRtsStatus( daemonId, eventInfo );
			}
			continue;
		}

		int result = 0;

		switch( eventInfo.eventKind )
		{
		case RTEventInfo::EventPresence:
			result = evtPresence( daemonId, *daemonInfo, eventInfo );
			break;

		case RTEventInfo::EventReference:
			result = evtReference( daemonId, *daemonInfo, eventInfo );
			break;

		case RTEventInfo::EventMsgIn:
			result = evtMessageIn( daemonId, *daemonInfo, eventInfo );
			break;

		case RTEventInfo::EventChain:
		case RTEventInfo::EventState:
		case RTEventInfo::EventFSM:
		case RTEventInfo::EventLateFSM:
			result = evtChain( daemonId, *daemonInfo, eventInfo );
			break;

		case RTEventInfo::EventMsgOut:
			result = evtMessageOut( daemonId, *daemonInfo, eventInfo );
			break;

		case RTEventInfo::EventMsgOutIn:
		{
			RTEventInfo senderEventInfo( RTEventInfo::EventMsgOutIn,
					eventInfo.message->_sendProbeId, -1 );

			senderEventInfo.actor = eventInfo.actor;
			senderEventInfo.info  = eventInfo.info;

			const RTDaemonInfo * senderDaemonInfo =
					theDaemon->eventMatches( senderEventInfo );

			if( senderDaemonInfo != nullptr )
			{
				result = evtMessageOutIn( daemonId,
						*senderDaemonInfo,
						*daemonInfo,
						eventInfo );
			}
			// Is it the initial transition?
			else if( eventInfo.message->sap() == nullptr )
			{
				RTDaemonInfo emptySenderDaemonInfo;

				result = evtMessageOutIn( daemonId,
						emptySenderDaemonInfo,
						*daemonInfo,
						eventInfo );
			}

			break;
		}

		case RTEventInfo::EventTimer:
		{
			RTDaemonInfo emptyDaemonInfo;

			// Is this a 'faked' call to informIn/At/Every?
			if( eventInfo.message->sap() == nullptr )
				result = evtMessageOutIn( daemonId,
						*daemonInfo,     // sender info
						emptyDaemonInfo, // bogus receiver
						eventInfo );
			else //  ordinary timeout event
				result = evtMessageOutIn( daemonId,
						emptyDaemonInfo, // bogus sender
						*daemonInfo,     // receiver info
						eventInfo );

			break;
		}

		case RTEventInfo::EventVariable:
			result = evtVariable( daemonId, *daemonInfo, eventInfo );
			break;

		case RTEventInfo::EventHistory:
			result = evtHistory( daemonId, *daemonInfo, eventInfo );
			break;

		case RTEventInfo::EventMsgQueue:
			result = evtMessageQueue( daemonId, *daemonInfo, eventInfo );
			break;

#ifdef _lint
		case RTEventInfo::EventControl:
#endif
		default:
			break;
		}

		if( result < 0 )
			return;

		shouldHalt |= result;

		if( shouldHalt != 0 )
			haltByProbe = 1;

		// If we're halting, send a halted message to the daemon.
		if( theDaemon->shouldHalt() != 0 )
		{
			eventInfo.targetMode = RTEventInfo::Halted;
			(void)evtRtsStatus( daemonId, eventInfo );
		}
	}
}
