/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#include <RTController.h>
#include <RTDiag.h>
#include <RTMessage.h>
#include <RTTimerActor.h>
#include <RTTimerId.h>
#include <RTTimerNode.h>
#include <RTTiming.h>

RTMessage::DeliverResult RTTimerNode::_deliver( void )
{
	// cancel must not delete this (repeating) timeout
	active = 1;

	RTMessage    * thisMsg  = msg;
	RTMessage    * nextMsg  = nullptr;
	RTTimerActor * timerActor;
	RTActor      * receiver = thisMsg->receiver();

	if( ( repeatInterval.tv_sec | repeatInterval.tv_nsec ) == 0L )
	{
		// invalidate all RTTimerId objects
		// that refer to this (one-shot) timeout
		_disconnectAll();
		timerActor = nullptr;
	}
	else if( (timerActor = RTTimerActor::soleInstance()) == nullptr )
	{
		owner->setError( RTController::noConnect,
				RTController::TimerInform,
				sap, nullptr );
	}
	else if( (nextMsg = owner->newMsg()) == nullptr )
	{
		owner->setError( RTController::noMem,
				RTController::SystemService, nullptr, nullptr );
	}
	else
	{
		buildMsg( nextMsg, receiver, thisMsg->getPriority() );
	}

	msg = nextMsg;

#if USE_THREADS
	critical.leave();
#endif

	if( nextMsg != nullptr )
		timerActor->informAgain( this );

	thisMsg->clear();
	thisMsg->setSignal( Timing::Base::rti_timeout );
	thisMsg->setData( data, type );
	thisMsg->setDestination( receiver, sap, 0 );

	RTMessage::DeliverResult result = thisMsg->deliver();

#if USE_THREADS
	critical.enter();
#endif

	active = 0;

	if( msg == nullptr )
		owner->freeTimerNode( this );

	return result;
}

RTMessage::DeliverResult RTTimerNode::deliver( const RTMessage * message )
{
	RTMessage::DeliverResult result;

	RTAssert( message->getType() == &RTType_RTTimerId );

#if USE_THREADS
	critical.enter();
#endif

	RTTimerId   * id   = static_cast<RTTimerId *>(message->data);
	RTTimerNode * node = id->node;

	if( node == nullptr ) // cancelled
		result = RTMessage::Accepted;
	else
		result = node->_deliver();

#if USE_THREADS
	critical.leave();
#endif

	return result;
}
