/*
 * Licensed Materials - Property of HCL and/or IBM
 * Copyright HCL Technologies Ltd. 2016, 2021. All Rights Reserved.
 * Copyright IBM Corporation 1999, 2016. All Rights Reserved.
 *
 * U.S. Government Users Restricted Rights -  Use, duplication or 
 * disclosure restricted by GSA ADP Schedule.
 */

#include <RTActor_class.h>
#include <RTAsciiEncoding.h>
#include <RTDaemon.h>
#include <RTDaemonInfo.h>
#include <RTEventInfo.h>
#include <RTFormat.h>
#include <RTMemoryUtil.h>
#include <RTMessage.h>
#include <RTPortDescriptor.h>
#include <RTPriority.h>
#include <RTToolsetObserver.h>
#include <RTTruncatingBuffer.h>

static const char _system_string[] = "<system>";

static char * getPortName( const RTActor_class & info, int portId, int index )
{
	const RTPortDescriptor * port = info.ports;

	for( int count = info.numPorts; --count >= 0; ++port )
	{
		if( port->id == portId )
		{
			int    portLength = RTMemoryUtil::strlen( port->name );
			char * tempPort   = new char[ static_cast<unsigned>(portLength) + 15U ];

			if( tempPort == nullptr )
				break;

			RTMemoryUtil::memcpy( tempPort, port->name, portLength );
			tempPort[ portLength ] = ':';
			if( RTFormat::_int( index, &tempPort[ portLength + 1 ], 14 ) == 0 )
			{
				delete [] tempPort;
				break;
			}

			return tempPort;
		}
	}

	return nullptr;
}

int RTToolsetObserver::evtMessageOutIn( int daemonId,
		const RTDaemonInfo & sndDaemonInfo,
		const RTDaemonInfo & recDaemonInfo,
		const RTEventInfo  & eventInfo )
{
	const RTDaemon * theDaemon = static_cast<const RTDaemon *>(daemons[ daemonId ]);

	if( theDaemon == nullptr )
		return 0;

	int shouldHalt = theDaemon->shouldHalt();

	if( theDaemon->shouldReport() == 0 )
		return shouldHalt;

	TO_evtMsgOutIn toSendData;
	TO_daemonInfo  toDaemonInfo;

	toDaemonInfo.number   = daemonId;
	toDaemonInfo.invokeID = nullptr;

	// send time
	toSendData.sendTime.sec  = eventInfo.message->_sendTime.tv_sec;
	toSendData.sendTime.nsec = eventInfo.message->_sendTime.tv_nsec;

	// create the send actor string
	if( sndDaemonInfo.info != nullptr )
	{
		toSendData.sendActor = buildString( sndDaemonInfo.actorPath,
				sndDaemonInfo.actorIndex );
		toSendData.sendPort  = getPortName( *sndDaemonInfo.info,
				eventInfo.message->_sendPortId,
				eventInfo.message->_sendPortIndex );
	}
	else
	{
		toSendData.sendActor = RTMemoryUtil::strdup( _system_string );
		toSendData.sendPort  = RTMemoryUtil::strdup( _system_string );
	}

	toSendData.sendMethod = eventInfo.sendMethod;

	// rec time
	toSendData.recTime.sec  = eventInfo.timestamp.tv_sec;
	toSendData.recTime.nsec = eventInfo.timestamp.tv_nsec;

	const char * signal;

	// create the rec actor string
	if( recDaemonInfo.info != nullptr )
	{
		toSendData.recActor = buildString( recDaemonInfo.actorPath,
				recDaemonInfo.actorIndex );
		signal = eventInfo.signalName;
	}
	else // a faked message caused by a call to informIn/At/Every
	{
		toSendData.recActor   = RTMemoryUtil::strdup( _system_string );
		toSendData.sendMethod = RTEventInfo::SendInvoke;

		switch( eventInfo.sendMethod )
		{
		case RTEventInfo::SendInformIn:    signal = "informIn";    break;
		case RTEventInfo::SendInformAt:    signal = "informAt";    break;
		case RTEventInfo::SendInformEvery: signal = "informEvery"; break;
#ifdef _lint
		case RTEventInfo::SendNormal:
		case RTEventInfo::SendInvoke:
		case RTEventInfo::SendReply:
#endif
		default:                           signal = "?";           break;
		}
	}

	toSendData.recPort  = eventInfo.portName;

	// create the signal/prio/data part
	char                 * msgDataBuf = new char [ msgDataBufSize ];
	const RTObject_class * type;

	if( eventInfo.message->data == nullptr )
		type =nullptr;
	else
		type = eventInfo.message->getType();

	if( type ==nullptr )
		msgDataBuf[ 0 ] = '\0';
	else
	{
		RTTruncatingBuffer buffer( msgDataBuf, msgDataBufSize );
		RTAsciiEncoding    coding( &buffer );

		(void)coding.put( eventInfo.message->data, type );
		(void)buffer.write( "", 1 );
	}

	toSendData.message.signal = signal;
	toSendData.message.prio   = eventInfo.message->getPriority();
	toSendData.message.data   = msgDataBuf;

	if( TO_encodeEvtMsgOutIn( &txBuffer, &toDaemonInfo, &toSendData )
			== ENC_SUCCESS )
	{
		sendTxBuffer();
	}

	delete [] msgDataBuf;
	delete [] toSendData.recActor;
	delete [] toSendData.sendPort;
	delete [] toSendData.sendActor;

	return shouldHalt;
}
